#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
[file name]
To do:
    [Write the tasks to do in this script]
-----------------    
Created on Fri May  1 09:12:12 2020

@author: Xiaodong Ming
"""
import pickle
import interp_line2surface as itp
import numpy as np
import matplotlib.pyplot as plt
from matplotlib import path
#% Create a river channel
xmax0 = 50
xmax1 = xmax0-5
xmin0 = 0
xmin1 = xmin0+5
x0 = np.arange(0, xmax0)
y0 = np.sin(x0/5)*5+10
x1 = np.arange(xmin1, xmax0-5)
y1 = np.sin(x1/5)*5
line_t = np.c_[x0, y0] # top line
line_b = np.c_[x1, y1] # bottom line
cellsize = 0.1
xy_cells = np.meshgrid(np.arange(xmin0, xmax0, cellsize),
                       np.arange(y1.min(), y0.max(), cellsize))
del x0, y0, x1, y1, xmax0, xmin0, xmax1, xmin1

#%% Discretize river channel to polygons
max_error = 0.1 # maximum error in meter
river_polys, break_points_b, break_points_t = \
    itp.discretize_river_channel(line_t, line_b, max_error)
#% show discretized polygons
# itp.show_discretized_channel(line_t, line_b, max_error)
#% show bank discretization
itp.show_discretized_bank(line_b, max_error)
itp.show_discretized_bank(line_t, max_error)
#% show projections of break points
itp.show_projected_break_points(break_points_t, line_b, color='b')
itp.show_projected_break_points(break_points_b, line_t, color='r')
#% show midlle points of the division lines
mid_points = (break_points_b+break_points_t)/2
plt.plot(mid_points[:,0], mid_points[:,1], ':')
plt.gca().set_aspect('equal', adjustable='box')
fig = plt.gcf()
plt.show()
# fig.savefig('bank_break_points', dpi=300)

#%% Show relative coords of gridded points
# compare two methods of converting coords

xy_cells_vec = np.c_[xy_cells[0].flatten(), xy_cells[1].flatten()]
# get bank-cross coordiantes, two ways
bc_pos = itp.convert_coords_2d(xy_cells_vec, line_b, line_t, 
                            max_error, itp.get_relative_position)
bc_dis = itp.convert_coords_2d(xy_cells_vec, line_b, line_t, 
                            max_error, itp.get_relative_distance)

fig, ax = plt.subplots(2, 1)

points_b = bc_pos[:,0]
points_c = bc_pos[:,1]
points_c[points_c>0.5] = 1-points_c[points_c>0.5]
points_z = np.round(points_c*10)
grids_z = points_z.reshape(xy_cells[0].shape)
ax[0].contourf(xy_cells[0], xy_cells[1], grids_z, 10, cmap='YlGnBu', alpha=1)
points_z = np.round(points_b*20)
grids_z = points_z.reshape(xy_cells[0].shape)
ax[0].contourf(xy_cells[0], xy_cells[1], grids_z, 20, cmap='Reds', alpha=0.5)
ax[0].set_aspect('equal', adjustable='box')

points_b = bc_dis[:,0]
points_c = bc_dis[:,1]
points_c[points_c>0.5] = 1-points_c[points_c>0.5]
points_z = np.round(points_c*10)
grids_z = points_z.reshape(xy_cells[0].shape)
ax[1].contourf(xy_cells[0], xy_cells[1], grids_z, 10, cmap='YlGnBu', alpha=1)
points_z = np.round(points_b*20)
grids_z = points_z.reshape(xy_cells[0].shape)
ax[1].contourf(xy_cells[0], xy_cells[1], grids_z, 20, cmap='Reds', alpha=0.5)
ax[1].set_aspect('equal', adjustable='box')
plt.tight_layout()
# fig.savefig('two_coords_interp_compare', dpi=300)
"""

The conclusion of this part is that get_relative_position can better convert
xy coords to bc coords.

"""
#%%============= show points within a channel polygon
x0 = np.arange(0,15)
y0 = np.sin(x0/10)*5+10
x1 = np.arange(5,15)
y1 = np.sin(x1/10)*5

line_t = np.c_[x0, y0] # top line
line_b = np.c_[x1, y1] # bottom line
point00 = line_b[0]
point01 = line_b[-1]
point11 = line_t[-1]
point10 = line_t[0]
grids_xy = np.meshgrid(np.arange(0, 15, 0.2), np.arange(0, 15, 0.2))
points_xy = np.c_[grids_xy[0].flatten(), grids_xy[1].flatten()]
section_poly = path.Path([point00, point01, point11, point10])
ind = section_poly.contains_points(points_xy)
points_xy = points_xy[ind, :]
#% p.contains_points([(.5, .5)])
line_cros0 = np.array([point00, point10])
line_cros1 = np.array([point01, point11])
plt.plot(x0, y0)
plt.plot(x1, y1)
plt.plot(line_cros0[:,0], line_cros0[:,1], ':')
plt.plot(line_cros1[:,0], line_cros1[:,1], ':')
plt.scatter(points_xy[:,0], points_xy[:,1], s=1)
#plt.annotate('O', point_inters)
plt.gca().set_aspect('equal', adjustable='box')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Created on Thu Mar 25 22:19:33 2021
# @author: Xiaodong Ming

"""
coords_interp
===========
To do:
    [objective of this script]

"""
import numpy as np
from numpy.linalg import norm

class line_1d_2d:
    def __init__(self, polyline):
        polyline = remove_duplicate_rows(polyline)
        # polyline = polyline[:, :2]
        p0 = polyline[:-1]
        p1 = polyline[1:]
        self.xy = polyline
        self.num_vert = polyline.shape[0] #number of vertice
        self.lengths = norm(p0-p1, axis=1)
        self.dx = p1[:, 0]-p0[:, 0]
        self.dy = p1[:, 1]-p0[:, 1]
        self.length_total = self.lengths.sum()
        d = self.lengths.cumsum()/self.length_total
        self.d = np.insert(d, 0, 0)
        self.dd = self.d[1:]-self.d[:-1]
    
    def xy_to_1d(self, xy_points):
        if xy_points.size == 2:
            xy_points = [xy_points]
        d_points = []
        for xy0 in xy_points:
            ind_n = nearest_points(xy0, self.xy, num=2)
            ind_0 = ind_n.min()
            verts = self.xy[ind_0:ind_0+2, :]
            point_prj, pos = point2segment(xy0, verts)
            if not pos:
                ind_0 = ind_n.max()
            verts = self.xy[ind_0:ind_0+2, :]
            d2 = norm(verts-xy0, axis=1)
            d_r = d2[0]/d2.sum()
            d0 = self.d[ind_0] + d_r*(self.d[ind_0+1]-self.d[ind_0])
            d_points.append(d0)
        d_points = np.array(d_points)
        return d_points
    
    def d_to_xy(self, d_points):
        #d>=0 and d<=1
        d_points = np.array(d_points)
        # if d_points.size == 1:
        #     d_points = [d_points]
        xy_list = []
        for n in np.arange(d_points.size):
            d0 = d_points[n]
            if d0 <= 0:
                x = self.xy[0, 0]
                y = self.xy[0, 1]
            elif d0>=1:
                x = self.xy[-1, 0]
                y = self.xy[-1, 1]
            else:
                ind  = np.where(self.d >= d0)[0]
                ind = ind[0]-1
                d_ratio = (d0-self.d[ind])/self.dd[ind]
                x = self.xy[ind, 0]+d_ratio*self.dx[ind]
                y = self.xy[ind, 1]+d_ratio*self.dy[ind]
            xy_list.append([x, y])
        xy = np.array(xy_list)
        if xy.size == 2:
            xy = xy.reshape(1,2)
        return xy
        
def get_line_coefs(p0, p1):
    """ Get line coefs of two points
    p1/p2: [2*N array] x and y coordinates 
    A * x + B * y = C
    """
    if p0.size==2:
        p0 = p0.reshape(1,2)
    if p1.size==2:
        p1 = p1.reshape(1,2)
    A = p0[:, 1] - p1[:, 1]
    B = p1[:, 0] - p0[:, 0]
    C = p1[:, 0]*p0[:, 1] - p1[:, 0]*p0[:, 1]
    return A, B, C

def remove_duplicate_rows(data_array):
    """ Remove duplicated points and preserve its order
    """
    u, ind = np.unique(data_array, axis=0, return_index=True)
    array_u = u[np.argsort(ind)]
    return array_u

def nearest_points(point_xy, poly_points, num=1):
    """
    find the nearest point on a line to a given point 
    Returns
    -------
    ind : index of the first nearest point.
    """
    distance = norm(poly_points-point_xy, axis=1)
    ind_sort = distance.argsort()
    if num==1:
        ind_nearest = ind_sort[0]
    else:
        ind_nearest = ind_sort[:num]
    return ind_nearest

def point2segment(points, segment):
    """ project point(s) to a line segment and identify its position  
    points: nx2 array
    segment: 2X2 array
    -------
    point_prj : nx2 array
        coordinates of the projected point.
    position: True|False
        indicate whether the projected point is in/outside the line segment.
        True: inside, False: outside 
    """
    p0 = segment[0]
    p1 = segment[1]
    line_vect = np.array(p0)-np.array(p1)
    if norm(line_vect)==0:
        unit_vect = line_vect*0
    else:
        unit_vect = line_vect/norm(line_vect)
    dot_values = np.dot(points-p0, unit_vect)
    if dot_values.size > 1:
        point_prj = p0+unit_vect*dot_values[:, None]
    else:
        point_prj = p0+unit_vect*dot_values
        point_prj = point_prj.reshape(1,2)
    dis2end = np.c_[norm(point_prj-p0, axis=1), norm(point_prj-p1, axis=1)]
    
    position = np.max(dis2end, axis=1) <= norm(p1-p0)
    if points.size == 2:
        position = position[0]
    return point_prj, position


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Created on Wed Mar 24 00:08:04 2021
# @author: Xiaodong Ming

"""
demo_functions
===========
To do:
    [objective of this script]

"""
import matplotlib.pyplot as plt
import numpy as np
#%% display functions
# def show_discretized_bank(line_bank, max_error=5):
#     """ Plot discretized bank sections
#     """
#     break_ind, break_points = discretize_bank_line(line_bank, max_error)
#     plt.plot(line_bank[:,0], line_bank[:,1], 'b', linewidth=0.5)
#     plt.plot(break_points[:,0], break_points[:,1], '.')
#     for n in np.arange(1, break_points.shape[0]):
#         section_line = break_points[[n-1, n]]
#         plt.plot(section_line[:,0], section_line[:,1], 'g--')
#     plt.gca().set_aspect('equal', adjustable='box')
#     return break_ind

# def show_projected_break_points(points_xy, break_points, color='r'):
#     """ Show break points on one bank projecting to the other bank
#     """
#     points_prj2bank = []
#     for one_point in points_xy:
#         point_prj, _ = project_point2polyline(break_points,
#                                                           one_point)
#         points_prj2bank.append(point_prj)
#     points_prj2bank = np.array(points_prj2bank)
#     plt.plot(points_prj2bank[:,0], points_prj2bank[:, 1], '*', alpha=0.4)
#     prj_vectors = points_prj2bank - points_xy
#     plt.quiver(points_xy[:, 0], points_xy[:, 1], prj_vectors[:, 0],
#                prj_vectors[:, 1], scale_units='xy', scale=1, width=0.003,
#                color=color, alpha=0.4)

#%%
# fig, ax = plt.subplots(1)
# n = 0
# for n in np.arange(len(river_sections)):# 
#     channel_dict = river_sections[n]
#     ax.plot(channel_dict['bank0'][:, 0], channel_dict['bank0'][:, 1], 
#             marker='o') #, c='b'
#     ax.plot(channel_dict['bank1'][:, 0], channel_dict['bank1'][:, 1], 
#             marker='o') #, c='b'
#     ax.plot(channel_dict['cross0'][:, 0], channel_dict['cross0'][:, 1], 
#             marker='.') #, c='r'
#     ax.plot(channel_dict['cross1'][:, 0], channel_dict['cross1'][:, 1], 
#             marker='.')
#     all_points = channel_dict['polygon']
#     # ax.plot(all_points[:, 0], all_points[:, 1], marker='.', c='r')
#     ax.annotate(str(n), xy=np.median(all_points, axis=0))
# ax.set_aspect('equal', adjustable='box')

#%% plot processed banklines and crosslines
def draw_arrow(ax, line_data, width=0.2, color='k', len_n=1):
    x_a = line_data[-1, 0]
    y_a = line_data[-1, 1]
    dx_a = line_data[-1, 0]-line_data[-2, 0]
    dy_a = line_data[-1, 1]-line_data[-2, 1]
    ax.arrow(x_a, y_a, dx_a*len_n, dy_a*len_n, width=width, color=color)

def draw_banklines_crosslines(bl0, bl1, crossline_list, fname=None):
    """ Plot all banklines and corsslines
    """
    fig1, ax1 = plt.subplots(1)
    ax1.plot(bl0[:, 0], bl0[:, 1], 'r', marker='.')
    draw_arrow(ax1, bl0, width=1, color='b')
    ax1.plot(bl1[:, 0], bl1[:, 1], 'r', marker='.')
    draw_arrow(ax1, bl1, width=1, color='b')
    
    for n in np.arange(len(crossline_list)):
        line_data = crossline_list[n]
        if line_data.size>0:
            ax1.plot(line_data[:, 0], line_data[:, 1], marker='.')
            # draw_arrow(ax1, line_data, width=1, len_n=5)
            if n%2:
                xy = line_data[0][:2]
            else:
                xy = np.mean(line_data[:, :2], axis=0)
            ax1.annotate(str(n), xy=xy)

    if fname is not None:
        ax1.set_aspect('equal', adjustable='box')
        fig1.savefig(fname, bbox_inches='tight', pad_inches=0, dpi=200)
    

#%%

# def show_discretized_channel(line_top, line_bottom, max_error):
#     """ Show discretized channel with polygons
#     call function discretize_river_channel
#     """
#     river_polys, _, _ = discretize_river_channel(line_top, line_bottom,
#                                                  max_error)
#     for geom in river_polys.geoms:
#         xs, ys = geom.exterior.xy
#         plt.fill(xs, ys, alpha=0.5)
#     plt.gca().set_aspect('equal', adjustable='box')
    
# def make_synthetic_crossline(cs, zs, p0=None, p1=None, num_of_stair=100):
#     """
#     make synthetic cross line using 1d relative coords of points
#     The number of points must be no less than 3. All these points must be
#     sorted and on the same straight line. 2-degree polyfit for 3 points, or
#     cubic interp1d for more than 3 points. If a start point(p0) and a end 
#     point(p1) are given, the function will return a 3-column array for xyz
#     coords. Otherwise, the return value will be a 2-column array for relative
#     coords and elevation

#     Parameters
#     ----------
#     cs : 1-col numpy array
#         1d relative coords.
#     zs : 1-col numpy array
#         elevation values.
#     num_of_stair: scalar
#         number of stairs (points) to the output cross line
#     p0 : xy coords of the start point on line section.
#     p1 : xy coords of the end point on line section.

#     Returns
#     -------
#     cz: 2-col numpy array
#         relative_position and z values of cross line points
#     xyz: 3-col numpy array for xyz coordiante of the cross line points
#         or 1-col numpy array only for z values.

#     """
#     cs_stairs = np.linspace(0, 1, num=num_of_stair+1, endpoint=True)
#     if zs.size >= 3:
#         line_fit = interpolate.interp1d(cs.flatten(), zs.flatten(),
#                                         kind='quadratic')
#     else:
#         raise ValueError('The number of points must be no less than 3')
#     z_stairs = line_fit(cs_stairs)
#     if (p0 is not None) and (p1 is not None):
#         x_stairs, y_stairs = get_xy_from_relative_position(p0, p1, cs_stairs)
#         xyz = np.c_[x_stairs, y_stairs, z_stairs]
#     else:
#         xyz = z_stairs
#     cz = np.c_[cs_stairs, z_stairs]
#     return cz, xyz
"""
#% Create a artificial river channel and cross section lines
# create bank lines
import pickle
xmax0 = 50
xmax1 = xmax0-5
xmin0 = 0
xmin1 = xmin0+5
x_bank_1 = np.arange(0, xmax0) # top line
y_bank_1 = np.sin(x_bank_1/5)*5+20
x_bank_0 = np.arange(xmin1, xmax0-5) # bottom line
y_bank_0 = np.sin(x_bank_0/5)*5+10
bankline0 = np.c_[x_bank_0, y_bank_0]
bankline1 = np.c_[x_bank_1, y_bank_1]
extent = (xmin0, xmax0, y_bank_0.min(), y_bank_1.max())
del x_bank_1, y_bank_1, x_bank_0, y_bank_0, xmax0, xmin0, xmax1, xmin1


# # upstream cross line

num_of_stair = 20
p0 = bankline0[0]
p1 = bankline1[0]
cs_given0 = np.array([0, 0.1, 0.3, 0.5, 1])
z_given0 = np.array([3, 2, 0, 1, 3])+5
_, cross_line0 = itp.make_synthetic_crossline(cs_given0, z_given0, p0, p1,
                                                      num_of_stair)
p0 = (14, 10)
p1 = (20, 17)
cs_given0 = np.array([0, 0.25, 0.5, 0.6, 0.9, 1])
z_given0 = np.array([3, 2, 1, 0, 2, 3])-2+5
_, cross_line1 = itp.make_synthetic_crossline(cs_given0, z_given0, p0, p1,
                                                      num_of_stair)


p0 = (30, 20)
p1 = (36.5, 14)
cs_given1 = np.array([0, 0.2, 0.5, 0.7, 1])
z_given1 = np.array([0,-2, -3, -2, 0])-0.5+5                          
_, cross_line2 = itp.make_synthetic_crossline(cs_given1, z_given1, p0, p1,
                                                      num_of_stair)

p0 = bankline0[-1]-0.5
p1 = bankline1[-1]+np.array([0.5, -1])
cs_given1 = np.array([0, 0.3, 0.6, 0.8, 1])
z_given1 = np.array([0,-2, -3, -2, 0])-2+5                      
_, cross_line3 = itp.make_synthetic_crossline(cs_given1, z_given1, p0, p1,
                                                      num_of_stair)

bankline1 = np.flipud(bankline1)
crossline_list = [cross_line0, cross_line1, cross_line2, cross_line3]
crossline_list = crossline_list[::-1]
header = {'ncols':120, 'nrows':60, 'cellsize':0.5, 
          'xllcorner':-5, 'yllcorner':0, 'NODATA_value':-9999}

sample_data = {'bankline0':bankline0, 'bankline1':bankline1,
                'crosslines':crossline_list, 'header':header}

with open('sample_data.pkl', 'wb') as f:
    pickle.dump(sample_data, f, pickle.HIGHEST_PROTOCOL)
"""


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Created on Mon Mar 29 16:05:14 2021
# @author: Xiaodong Ming

"""
convert_coords
===========
To do:
    [objective of this script]

"""
import numpy as np
from matplotlib import path
from numpy.linalg import norm
from scipy import interpolate
from hipims_io import Raster
from rasterio import mask
from .channel_geometry import discretize_river_section
from .channel_geometry import get_intersection, distance_p2l, break_bankline
from .channel_geometry import preprocessing_data, split_channel
from .channel_geometry import construct_channel_polygon, in_channel
from .channel_geometry import get_bounding_box
from .channel_geometry import match_break_points_by_relative_pos

#%%
def cross_section2grid_elevation(bankline0, bankline1, crossline_list,
                                 cellsize, bounding_box=None, max_error=None):
    # preprocess data: sort and trim
    bl0, bl1, cl_sort = preprocessing_data(bankline0, bankline1, crossline_list)
    # split river polygon using crosslines
    river_sections = split_channel(cl_sort, bl0, bl1)
    river_channel, _, _ = construct_channel_polygon(bl0, bl1, cl_sort[0], 
                                              cl_sort[-1])
    # create a river topography grid
    if bounding_box is None:
        bounding_box = get_bounding_box(river_channel)
    x_grid, y_grid, header = generate_cells_xy(bounding_box, cellsize)
    if max_error is None:
        max_error = cellsize*2
    x_valid, y_valid, ind_valid = extract_valid_xy(x_grid, y_grid, header, 
                                                   river_channel)
    z_valid = x_valid+np.nan
    n = 0
    for channel_dict in river_sections:
        bs, cs, zs = interp_one_reach(channel_dict, x_valid, y_valid, 
                                      max_error)
        ind = ~np.isnan(zs)
        z_valid[ind] = zs[~np.isnan(zs)]
        print(str(n)+'/'+str(len(river_sections)))
        n = n+1
    z_grid = x_grid+np.nan
    z_grid[ind_valid] = z_valid
    return z_grid, header

def generate_cells_xy(bounding_box, cellsize):
    """get xy coordiantes of the center of each cell 
    bounding_box: 2x2 array [[xmin, ymin], [xmax, ymax]] 
    cellsize: length of each cell
    """
    xmin = np.min(bounding_box[:, 0])
    xmax = np.max(bounding_box[:, 0])
    ymin = np.min(bounding_box[:, 1])
    ymax = np.max(bounding_box[:, 1])
    xllcorner = xmin-cellsize/2
    yllcorner = ymin-cellsize/2
    ncols = np.abs((xmax-xmin)/cellsize).astype('int')
    nrows = np.abs((ymax-ymin)/cellsize).astype('int')
    header = {'ncols':ncols,
              'nrows':nrows,
              'xllcorner':xllcorner, 
              'yllcorner':yllcorner,
              'cellsize':cellsize, 
              'NODATA_value':-9999}
    x00 = xllcorner+cellsize/2
    y00 = yllcorner+nrows*cellsize-cellsize/2
    x_vect = np.linspace(x00, x00+cellsize*ncols, num=ncols, endpoint=False)
    y_vect = np.linspace(y00, y00-cellsize*nrows, num=nrows, endpoint=False)
    x_grid, y_grid = np.meshgrid(x_vect, y_vect)
    return x_grid, y_grid, header

def extract_valid_xy(x_grid, y_grid, header, river_channel):
    # extract valid xy coordinates and return as a 2-col array plus a ind array
    # river_channel: path or numpy arrays providing xy of river channel vertex
    raster_obj = Raster(array=x_grid, header=header)
    ind_valid = rasterise_polygon(river_channel, raster_obj)
    x_grid_1d = x_grid[ind_valid]
    y_grid_1d = y_grid[ind_valid]
    return x_grid_1d, y_grid_1d, ind_valid

def rasterise_polygon(river_channel, raster_obj):
    # river_channel is a Path object
    # raster_obj: a Raster object
    if hasattr(river_channel, 'vertices'):
        xy_list = river_channel.vertices.tolist()
    else:
        xy_list = river_channel.tolist()
    shapes = [{'type':'Polygon', 
               'coordinates':[xy_list]}]
    ds_rio = raster_obj.to_rasterio_ds()
    out_image, _ = mask.mask(ds_rio, shapes) #, crop=True
    rasterized_array = out_image[0]
    rasterized_array[np.isnan(rasterized_array)] = ds_rio.nodata
    index_array = np.full(rasterized_array.shape, True)
    index_array[rasterized_array == ds_rio.nodata] = False
    ds_rio.close()
    return index_array

def interp_one_reach(channel_dict, x, y, max_error):
    """ interp xy points within one reach splitted by two cross-section lines
    Parameters
    ----------
    channel_dict : dictionary provide xy coords of two bank lines and two
        cross lines of one river reach.
    x, y : coords of xy cells within the reach
    max_error : maximum error to discritise the reach bank lines

    Returns
    -------
    bs : bank coords
    cs : cross coords
    zs : elevation values
    ind: index of valid cells within the reach 
    """
    cl0 = channel_dict['cross0']
    cl1 = channel_dict['cross1']
    bl0 = channel_dict['bank0']
    bl1 = channel_dict['bank1']
    xy_cells_vec = np.c_[x.flatten(), y.flatten()]
    b_vec = x.flatten()+np.nan
    c_vec = x.flatten()+np.nan
    z_vec = x.flatten()+np.nan
    ind = in_channel(channel_dict, xy_cells_vec, radius=0.00001)
    if ind.sum() > 0:
        xy_valid = xy_cells_vec[ind]
        b, c = xy2bc(xy_valid, bl0, bl1, max_error)
        z = interp_z_on_bc(b, c, cl0, cl1)
        b_vec[ind] = b
        c_vec[ind] = c
        z_vec[ind] = z
    bs = b_vec.reshape(x.shape)
    cs = c_vec.reshape(x.shape)
    zs = z_vec.reshape(x.shape)
    return bs, cs, zs

def xy2bc(points_xy, bl0_sec, bl1_sec, max_error,
          relative_pos=True):
    """Convert points_xy to bc coords defined within two river cross-sections
    points_xy: XY points to be converted
    line_bank0/1: points XY coords of two bank lines  between the crosslines
    max_error: maximum discretization error (map unit)
    relative_pos: yes to choose get_relative_position to convert xy to bc,
        otherwise, choose get_relative_distance
    one river section would be divided into several quadr-/tri- angle polygons
    coords transfer is processed within each polygon
    """
    points_bc = points_xy*0+np.nan
    _, bl0_broken = break_bankline(bl0_sec, max_error)
    _, bl1_broken = break_bankline(bl1_sec, max_error)
    if relative_pos:
        convert_fun = get_relative_position
    else:
        convert_fun = get_relative_distance
    bl0_broken_e, bl1_broken_e = match_break_points_by_relative_pos(bl0_broken, bl1_broken)
    river_polys = discretize_river_section(bl0_broken_e, bl1_broken_e)
    b_coords_div = get_b_coords_of_bank_divisions(bl0_broken_e, bl1_broken_e)
    n = 1
    for geom in river_polys.geoms:
        xs, ys = geom.exterior.xy
        point00 = bl0_broken_e[n-1]
        point01 = bl0_broken_e[n]
        point10 = bl1_broken_e[n-1]
        point11 = bl1_broken_e[n]
        polygon_n = path.Path([point00, point01, point11, point10])       
        # add a small radius to include points on edges
        ind = polygon_n.contains_points(points_xy, radius=0.00001)
        if ind.sum()>0:
            points_xy_in = points_xy[ind, :]
            b_coords, c_coords = convert_fun(points_xy_in,
                                          point00, point01, point10, point11)
            # convert local B coords to global
            b0 = b_coords_div[n-1]
            b1 = b_coords_div[n]
            b_coords = b_coords*(b1-b0)+b0
            points_bc[ind, :] = np.c_[b_coords, c_coords]
        n = n+1
    bs = points_bc[:, 0]
    cs = points_bc[:, 1]
    return bs, cs

def get_b_coords_of_bank_divisions(bl0_broken_e, bl1_broken_e):
    """ Get b coordinates of bank division lines within one river section
    """
    mid_points = (bl0_broken_e+bl1_broken_e)/2
    mid_gaps = mid_points[1:]-mid_points[:-1]
    mid_gaps = norm(mid_gaps, axis=1)
    mid_gaps_cumsum = np.cumsum(mid_gaps)
    b_coords = mid_gaps_cumsum/mid_gaps_cumsum.max()
    b_coords = np.insert(b_coords, 0, 0)
    return b_coords

def get_relative_distance(points_xy, point00, point01, point10, point11):
    """ Convert a Cartesian coordinates to a relative coordinates (B, C)
    points_xy: 2-col array for X and Y coordinates
    points position: point00, point01, point10, point11
    Return:
        coords_B: relative coordiantes along bank line
        coords_C: relative coordiantes along cross line
    """
    line_bank0 = np.array([point00, point01])
    line_bank1 = np.array([point10, point11])
    line_cros0 = np.array([point00, point10])
    line_cros1 = np.array([point01, point11])
    # axis along bank
    d2cros0 = distance_p2l(line_cros0, points_xy)
    d2cros1 = distance_p2l(line_cros1, points_xy)
    coords_B = d2cros0/(d2cros0+d2cros1)
    # axis along cross
    d2bank0 = distance_p2l(line_bank0, points_xy)
    d2bank1 = distance_p2l(line_bank1, points_xy)
    coords_C = d2bank0/(d2bank0+d2bank1)
    return coords_B, coords_C

def get_relative_position(points_xy, point00, point01, point10, point11):
    """ Find relative position of points_xy in a polygon
    via intersections of sides
    """
    line_bank0 = np.array([point00, point01])
    line_bank1 = np.array([point10, point11])
    line_cros0 = np.array([point00, point10])
    line_cros1 = np.array([point01, point11])
    ip_bank = get_intersection(line_bank0, line_bank1) #bank lines intersection
    ip_cros = get_intersection(line_cros0, line_cros1) #cross lines intersection
    # axis along bank
    if np.isnan(ip_bank.flat[0]): # parallel lines use projected distance
        d1 = distance_p2l(line_cros0, points_xy)
        d2 = distance_p2l(line_cros1, points_xy)
    else:
        d1, d2 = relative_pos_between2lines(points_xy, line_cros0, line_cros1,
                                            ip_bank)
    coords_B = d1/(d1+d2)
    # axis along cross section
    if np.isnan(ip_cros.flat[0]): # parallel lines use projected distance
        d1 = distance_p2l(line_bank0, points_xy)
        d2 = distance_p2l(line_bank1, points_xy)
    else:
        d1, d2 = relative_pos_between2lines(points_xy, line_bank0, line_bank1,
                                            ip_cros)
    coords_C = d1/(d1+d2)
    return (coords_B, coords_C)

def relative_pos_between2lines(points, side0, side1, intersection):
    """ Get relative position of points between two line sections
    sides1/sides2: two opposite sides of a quadrangle
    intersection：the intersection of the other two sides
    """
    point_num = points.shape[0] 
    intersection = np.array(intersection).flatten()
    intersection_rep = np.array([intersection,]*point_num)
    connect_lines = np.dstack([intersection_rep, points])
    connect_lines = connect_lines.transpose(2,1,0)
    if norm(side0[0]-side0[1])<=0.000001: # (nearly) triangle
        points_inters1 = np.array([side0[0].flatten()]*point_num)
    else:
        points_inters1 = get_intersection(connect_lines, side0)
        if np.array(points_inters1).size>2:
            points_inters1 = np.array(points_inters1).transpose(1,0)
    if norm(side1[0]-side1[1])<=0.000001: #(nearly) triangle
        points_inters2 = np.array([side1[0].flatten()]*point_num)
    else:
        points_inters2 = get_intersection(connect_lines, side1)
        if np.array(points_inters2).size>2:
            points_inters2 = np.array(points_inters2).transpose(1,0)
    d1 = norm(points_inters1-points, axis=1)
    d2 = norm(points_inters2-points, axis=1)
    return d1, d2

def interp_z_on_bc(bs, cs, cz0, cz1):
    """
    generate z values of relative bc coords according to cross line z values
    Parameters
    ----------
    bs : 1-col numpy array
        b(ank) coordinates.
    cs : 1-col numpy array
        c(ross) coordinates.
    cz0 : 2-col numpy array or 3-col numpy array(x,y,z)
        c(ross) and z coordinates of the start cross line.
    cz1 : 2-col numpy array or 3-col numpy array(x,y,z)
        c(ross) and z coordinates of the end cross line.

    Returns
    -------
    zs: 1-col numpy array
        interpolated z value of points defined by bs and cs
    """
    if cz0.shape[1]==3:
        cz = normalise_xy_coords(cz0[:, 0], cz0[:, 1])
        cz0 = np.c_[cz, cz0[:, 2]]
        cz = normalise_xy_coords(cz1[:, 0], cz1[:, 1])
        cz1 = np.c_[cz, cz1[:, 2]]
    c_interp0 = interpolate.interp1d(cz0[:, 0], cz0[:, 1], fill_value="extrapolate")
    c_interp1 = interpolate.interp1d(cz1[:, 0], cz1[:, 1], fill_value="extrapolate")
    z0 = c_interp0(cs)
    z1 = c_interp1(cs)
    zs = z0*(1-bs)+z1*bs
    return zs

def normalise_xy_coords(x, y):
    """ Convert cross line points to 1d relative coords cs
    convert xy coordiantes of points on one straignt line to relative coords
    and return values ranging 0 to 1, which cooresponding to position 
    starting from the first point and ending at the last point

    Parameters
    ----------
    x : 1-col numpy array
        x coords.
    y : 1-col numpy array
        y coords.

    Returns
    -------
    cs : 1-col numpy array
        1d relative coords ranging 0 to 1.

    """
    x.shape = (x.size, 1)
    y.shape = (y.size, 1)
    p0 = np.array([x[0], y[0]]).reshape((1,2)) # start point
    p1 = np.array([x[-1], y[-1]]).reshape((1,2)) # end point
    dist = np.linalg.norm(p1-p0, axis=1) # total length of the line
    d_points = np.linalg.norm(np.c_[x, y]-p0, axis=1) # distance to the start
    cs = d_points/dist
    return cs



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
interp_line2surface
To do:
    [Write the tasks to do in this script]

Input requirements:
    1. all river bank line points should be sorted towards the same flow 
        direction: upstream to downstream or the opposite
    2. all cross section points should be sorted from one bank to the other
    
General information:

    1. A b-c (bank-crossline) coordiante system:
        point10--------------bank1-------------point11
           |                                      |
           |                                      |
        cross0                                  cross1
           |                                      |
           |                                      |
        point00--------------bank0-------------point01

Universal variable names:
    line_poly: 2-col array, 1st col is x coords, 2nd col is y coords
        a polygonal line defined by a series of points (more than two)
    line_straight: 2x2 array, 1st col is x coords, 2nd col is y coords
        a straight line defined only by two points
    line_section: 2x2 array, 1st col is x coords, 2nd col is y coords
        a straight line between two points
    p0, p1, ..., pN: 2-element array provide x,y coords of one point
    points: 2-col array providing x(1st) and y(2nd) coords of multiple points
    
-----------------    
Created on Wed Apr 29 12:48:33 2020
@author: Xiaodong Ming
"""
#%%
import numpy as np
from numpy.linalg import norm
from matplotlib import path
from shapely import geometry
from scipy import interpolate
from channel_geometry import get_intersection, in_channel, trim_bankline
from channel_geometry import nearest_points
#%% 
def cross_section2grid_elevation(bankline0, bankline1, crossline_list,
                                 grid_header, max_error=None):
    
    # split river polygon using crosslines
    channel_list = split_river_polygon(bankline0, bankline1, crossline_list)
    # create a river topography grid
    cellsize = grid_header['cellsize']
    nrows = grid_header['nrows']
    ncols = grid_header['ncols']
    x00 = grid_header['xllcorner']+cellsize/2
    y00 = grid_header['yllcorner']+nrows*cellsize-cellsize/2
    x_vect = np.linspace(x00, x00+cellsize*ncols, num=ncols, endpoint=False)
    y_vect = np.linspace(y00, y00-cellsize*nrows, num=nrows, endpoint=False)
    x_grid, y_grid = np.meshgrid(x_vect, y_vect)
    if max_error is None:
        max_error = cellsize*2
    x_grid_1d = x_grid.flatten()
    y_grid_1d = y_grid.flatten()
    z_grid_1d = x_grid_1d+np.nan
    xy_1d = np.c_[x_grid_1d, y_grid_1d]
    for channel_dict in channel_list:
        ind = in_channel(channel_dict, xy_1d)
        xy_cells_vec = xy_1d[ind]
        if ind.sum() > 0:
            
            bc_cells = convert_coords_2d(xy_cells_vec, channel_dict['bank0'], 
                                         channel_dict['bank1'], max_error)
            bs = bc_cells[:, 0]
            cs = bc_cells[:, 1]
            zs = bc_interp_2d(bs, cs, channel_dict['cross0'],
                              channel_dict['cross1'])
            z_grid_1d[ind] = zs
    z_grid = np.reshape(z_grid_1d, x_grid.shape)
    return z_grid

#********************* Coordinates conversion ********************************
def get_split_lines_middle_coords(break_points_full_b, break_points_full_t):
    """ Get BC coordinates of channel split lines
    break_points_full_b/c: expanded break points from function 
        expand_break_points
    """
    mid_points = (break_points_full_b+break_points_full_t)/2
    mid_gaps = mid_points[1:]-mid_points[:-1]
    mid_gaps = norm(mid_gaps, axis=1)
    mid_gaps_cumsum = np.cumsum(mid_gaps)
    mid_bc_coords = mid_gaps_cumsum/mid_gaps_cumsum.max()
    mid_bc_coords = np.insert(mid_bc_coords, 0, 0)
    return mid_bc_coords

def get_relative_distance(points_xy, point00, point01, point10, point11):
    """ Convert a Cartesian coordinates to a relative coordinates (B, C)
    points_xy: 2-col array for X and Y coordinates
    points position: point00, point01, point10, point11
    Return:
        coords_B: relative coordiantes along bank line
        coords_C: relative coordiantes along cross line
    """
    line_bank0 = np.array([point00, point01])
    line_bank1 = np.array([point10, point11])
    line_cros0 = np.array([point00, point10])
    line_cros1 = np.array([point01, point11])
    # axis along bank
    d2cros0 = distance_p2l(line_cros0, points_xy)
    d2cros1 = distance_p2l(line_cros1, points_xy)
    coords_B = d2cros0/(d2cros0+d2cros1)
    # axis along cross
    d2bank0 = distance_p2l(line_bank0, points_xy)
    d2bank1 = distance_p2l(line_bank1, points_xy)
    coords_C = d2bank0/(d2bank0+d2bank1)
    return coords_B, coords_C

def get_relative_position(points_xy, point00, point01, point10, point11):
    """ Find relative position of points_xy in a polygon
    via intersections of sides
    """
    line_bank0 = np.array([point00, point01])
    line_bank1 = np.array([point10, point11])
    line_cros0 = np.array([point00, point10])
    line_cros1 = np.array([point01, point11])
    ip_bank = get_intersection(line_bank0, line_bank1) #bank lines intersection
    ip_cros = get_intersection(line_cros0, line_cros1) #cross lines intersection
    # axis along bank
    if np.isnan(ip_bank[0]): # parallel lines use projected distance
        d1 = distance_p2l(line_cros0, points_xy)
        d2 = distance_p2l(line_cros1, points_xy)
    else:
        d1, d2 = relative_position_1d(points_xy, line_cros0, line_cros1,
                                       ip_bank)
    coords_B = d1/(d1+d2)
    # axis along cross section
    if np.isnan(ip_cros[0]): # parallel lines use projected distance
        d1 = distance_p2l(line_bank0, points_xy)
        d2 = distance_p2l(line_bank1, points_xy)
    else:
        d1, d2 = relative_position_1d(points_xy, line_bank0, line_bank1,
                                       ip_cros)
    coords_C = d1/(d1+d2)
    return (coords_B, coords_C)

def relative_position_1d(points, sides1, sides2, intersection):
    """ Get relative position of points between two line sections
    sides1/sides2: two opposite sides of a quadrangle
    intersection：the intersection of the other two sides
    """
    point_num = points.shape[0] 
    intersection = np.array(intersection).flatten()
    intersection_rep = np.array([intersection,]*point_num)
    connect_lines = np.dstack([intersection_rep, points])
    connect_lines = connect_lines.transpose(2,1,0)
    if norm(sides1[0]-sides1[1])<=0.000001: # (nearly) triangle
        points_inters1 = np.array([sides1[0].flatten()]*point_num)
    else:
        points_inters1 = get_intersection(connect_lines, sides1)
        points_inters1 = np.array(points_inters1).transpose(1,0)
    if norm(sides2[0]-sides2[1])<=0.000001: #(nearly) triangle
        points_inters2 = np.array([sides2[0].flatten()]*point_num)
    else:
        points_inters2 = get_intersection(connect_lines, sides2)
        points_inters2 = np.array(points_inters2).transpose(1,0)
    d1 = norm(points_inters1-points, axis=1)
    d2 = norm(points_inters2-points, axis=1)
    return d1, d2

def normalise_xy_coords(x, y):
    """ Convert cross line points to 1d relative coords cs
    convert xy coordiantes of points on one straignt line to relative coords
    and return values ranging 0 to 1, which cooresponding to position 
    starting from the first point and ending at the last point

    Parameters
    ----------
    x : 1-col numpy array
        x coords.
    y : 1-col numpy array
        y coords.

    Returns
    -------
    cs : 1-col numpy array
        1d relative coords ranging 0 to 1.

    """
    x.shape = (x.size, 1)
    y.shape = (y.size, 1)
    p0 = np.array([x[0], y[0]]).reshape((1,2)) # start point
    p1 = np.array([x[-1], y[-1]]).reshape((1,2)) # end point
    dist = np.linalg.norm(p1-p0, axis=1) # total length of the line
    d_points = np.linalg.norm(np.c_[x, y]-p0, axis=1) # distance to the start
    cs = d_points/dist
    return cs

def get_xy_from_relative_position(p0, p1, relative_position):
    """ Get xy coords of points on a line section defined by relative position
    Parameters
    ----------
    p0 : xy coords of the start point on line section.
    p1 : xy coords of the end point on line section.
    relative_position : numpy array
        relative_position ranging from 0 to 1.

    Returns
    -------
    x : numpy array
        x coords of the new points.
    y : numpy array
        y coords of the new points.
    """
    x = p0[0]+(p1[0]-p0[0])*relative_position
    y = p0[1]+(p1[1]-p0[1])*relative_position
    return x, y

def convert_coords_2d(points_xy, line_bank0, line_bank1, max_error,
                   conver_fun=get_relative_position):
    """Convert points_xy to bc coords defined within two river cross-sections
    points_xy: XY points to convert
    line_bank0/1: points XY coords of two bank lines  between the crosslines
    max_error: maximum discretization error (map unit)
    """
    points_bc = points_xy*0+np.nan
    line_b = line_bank0
    line_t = line_bank1
    _, breaks_t, breaks_b = discretize_river_channel(line_t, line_b, max_error)
    mid_bc_coords = get_split_lines_middle_coords(breaks_t, breaks_b)
    for n in np.arange(1, breaks_t.shape[0]):
        point00 = breaks_t[n-1]
        point01 = breaks_t[n]
        point10 = breaks_b[n-1]
        point11 = breaks_b[n]
        section_poly = path.Path([point00, point01, point11, point10])       
        # add a small radius to include points on edges
        ind = section_poly.contains_points(points_xy, radius=0.00001)
        if ind.sum()>0:
            points_xy_in = points_xy[ind, :]
#        conver_fun = cartesian2relative #uniformed_relative_position #
            points_B, points_C = conver_fun(points_xy_in,
                                          point00, point01, point10, point11)
        # conver local B coords to global
            B0 = mid_bc_coords[n-1]
            B1 = mid_bc_coords[n]
            points_B = points_B*(B1-B0)+B0
            points_bc[ind, :] = np.c_[points_B, points_C]
    return points_bc

#********************* Construct river channel *******************************
def discretize_bank_line(line_bank, max_error):
    """discretize bank line to straight line sections
    line_bank: X and Y coordinates of bank line points
    max_error: the max distance between line points to straight bank line
    Return:
        break_ind: index of the break points
    """
    break_ind = []
    ind_0 = 0 # start index of a bank section
    ind_1 = line_bank.shape[0] # end index of a bank section
    while ind_0 < ind_1:
        section_line = np.array([line_bank[ind_0], line_bank[ind_1-1]])
        points_xy = line_bank[ind_0:ind_1]
        distance = distance_p2l(section_line, points_xy)
        if distance.max() > max_error:
            ind_1 = distance.argmax()+ind_0
            ind_1 = ind_1.astype('int64')
        else:
            break_ind.append(ind_1)
            ind_0 = ind_1
            ind_1 = line_bank.shape[0]
    break_ind = np.r_[0, np.array(break_ind)-1]
    break_points = line_bank[break_ind]
    return break_ind, break_points

def discretize_river_channel(line_top, line_bottom, max_error):
    """discretize river channel to 3/4-side polygons
    Return:
        river_polys: shapely multi-polygon
        break_points_full_b: full break points on line_bottom
        break_points_full_t: full break points on line_top
    """
    _, break_points_t = discretize_bank_line(line_top, max_error)
    _, break_points_b = discretize_bank_line(line_bottom, max_error)
    points_xy_t = break_points_t[1:-1] # line_top[1:-1] # 
    points_xy_b = break_points_b[1:-1] # line_bottom[1:-1] # 
    # expand break points with projections from break points on the other bank
    # break_points_full_b and break_points_full_t should have the same size
    break_points_full_b = expand_break_points(break_points_b, points_xy_t)
    break_points_full_t = expand_break_points(break_points_t, points_xy_b)
    polygon_list = []
    for n in np.arange(1, break_points_full_b.shape[0]):
        point00 = break_points_full_b[n-1]
        point01 = break_points_full_t[n-1]
        point11 = break_points_full_t[n]
        point10 = break_points_full_b[n]
        polygon = geometry.Polygon([point00, point01, point11, point10])
        polygon_list.append(polygon)
    river_polys = geometry.MultiPolygon(polygon_list)
    return river_polys, break_points_full_b, break_points_full_t

def project_point2polyline(polyline, point_xy):
    """
    Project a point to the two nearest line sections and return index

    Parameters
    ----------
    polyline : 2-col array
        xy coords of points of a polyline.
    point_xy : 2-element array
        xy coords of the point to be projected.

    Returns
    -------
    point_projected : 2-element array
        xy coords of the projected point.
    ind_insert : int scalar
        position index of the projected point on the
            polyline points array.

    """
    ind_nearest2 = nearest_points(point_xy, polyline, 2)
    if ind_nearest2[0] == 0 or ind_nearest2[0] == polyline.shape[0]-1:
        ind_nearest = ind_nearest2[1]
    else:
        ind_nearest = ind_nearest2[0]
    
    p0 = polyline[ind_nearest-1]
    p1 = polyline[ind_nearest]
    point_prj_0, out0 = project_point2line_section(p0, p1, point_xy)
    out0 = out0[0]
    p0 = polyline[ind_nearest]
    p1 = polyline[ind_nearest+1]
    point_prj_1, out1 = project_point2line_section(p0, p1, point_xy)
    out1 = out1[0]
    # select the best projected point
    if out0:
        if out1: # outside both two sections, select the nearest break point
            point_projected = polyline[ind_nearest]
            ind_insert = ind_nearest
        else: # project on the second section
            point_projected = point_prj_1
            ind_insert = ind_nearest+1
    else: # out0 is false
        if out1: # project on the first section
            point_projected = point_prj_0
            ind_insert = ind_nearest
        else: # onside both two sections, select the nearest projected point
            if norm(point_prj_0-point_xy) > norm(point_prj_1-point_xy):
                point_projected = point_prj_1
                ind_insert = ind_nearest+1
            else:
                point_projected = point_prj_0
                ind_insert = ind_nearest
    return point_projected, ind_insert


def expand_break_points(break_points, points_to_project):
    """ Expand bank break points with points projected from the other bank
    break_points: to define break points of bank line sections, 
        sequence mattters
    points_to_project: coordinates of points on the other bank
    Return:
        break_points_expand
    """
#    break_points_exp = break_points+0 # break points expanded
    point_prj_list = []
    ind_insert_list = []
    for one_point in points_to_project:
        point_prj, ind_insert = project_point2polyline(
                break_points, one_point)
        point_prj_list.append(point_prj.flatten())
        ind_insert_list.append(ind_insert)
    break_points_expand = np.insert(break_points, ind_insert_list,
                                    point_prj_list, axis=0)
    return break_points_expand

def split_river_polygon(bankline0, bankline1, crossline_list):
    """
    bankline: bankline points sorted from upstream to downstream
    crossline0, crossline1: crossline points

    """
    channel_list = []
    for n in np.arange(len(crossline_list)-1):
        cline0 = crossline_list[n]
        cline1 = crossline_list[n+1]
        bline0 = trim_bankline(bankline0, cline0, cline1)
        bline1 = trim_bankline(bankline1, cline0, cline1)
        channel_dict = {'bank0':bline0, 'bank1':bline1,
                        'cross0':cline0, 'cross1':cline1}
        channel_list.append(channel_dict)
    return channel_list
    
#***************** Interpolation and calculation *****************************

def bc_interp_2d(bs, cs, cz0, cz1):
    """
    generate z values of relative bc coords according to cross line z values
    Parameters
    ----------
    bs : 1-col numpy array
        b(ank) coordinates.
    cs : 1-col numpy array
        c(ross) coordinates.
    cz0 : 2-col numpy array or 3-col numpy array(x,y,z)
        c(ross) and z coordinates of the start cross line.
    cz1 : 2-col numpy array or 3-col numpy array(x,y,z)
        c(ross) and z coordinates of the end cross line.

    Returns
    -------
    zs: 1-col numpy array
        interpolated z value of points defined by bs and cs
    """
    if cz0.shape[1]==3:
        cz = normalise_xy_coords(cz0[:, 0], cz0[:, 1])
        cz0 = np.c_[cz, cz0[:, 2]]
        cz = normalise_xy_coords(cz1[:, 0], cz1[:, 1])
        cz1 = np.c_[cz, cz1[:, 2]]
    c_interp0 = interpolate.interp1d(cz0[:, 0], cz0[:, 1], fill_value="extrapolate")
    c_interp1 = interpolate.interp1d(cz1[:, 0], cz1[:, 1], fill_value="extrapolate")
    z0 = c_interp0(cs)
    z1 = c_interp1(cs)
    zs = z0*(1-bs)+z1*bs
    return zs


#%%*****************Geometry calculation *************************************


def two_lines_intersect(line0, line1):
    """ Check if two sections intersect
    line0, line2: 2X2 array, each row represents a point of line
    Return: True|False
    """
    line = geometry.LineString(line0)
    other = geometry.LineString(line1)
    return(line.intersects(other))
    
def distance_p2l(line_straight, points):
    """Calculate the distance from a point to a line defined by two points
    line_a: 2x2 array, 1st col is X and 2nd col is Y
    points: x and y coordinates of a point, can also be a 2-col array for 
        more than one points
    """
    p1 = line_straight[0, 0:2]
    p2 = line_straight[1, 0:2]
    if norm(p2-p1)==0:
        distance = norm(points-p1, axis=1)
    else:
        distance = np.cross(p2-p1, points-p1, axisb=1)/norm(p2-p1)
    distance = np.abs(distance)
    return distance

def project_point2line_section(p0, p1, points):
    """
    Project a point to a line section
    Returns
    -------
    point_prj : 2-element array
        coordinates of the projected point.
    outside : logical
        to indicate whether the projected point is outside the line section.

    """
    line_vect = np.array(p0)-np.array(p1)
    unit_vect = line_vect/norm(line_vect)
    dot_values = np.dot(points-p0, unit_vect)
    if dot_values.size > 1:
        point_prj = p0+unit_vect*dot_values[:, None]
    else:
        point_prj = p0+unit_vect*dot_values
        point_prj = point_prj.reshape(1,2)
    dis2end = np.c_[norm(point_prj-p0, axis=1), norm(point_prj-p1, axis=1)]
    outside = np.max(dis2end, axis=1)>norm(p1-p0)
    
    return point_prj, outside


def point_position_on_linesection(point, line_section):
    """
    Find the relative position of one point to a section of line which pass
    through one point
    
    Parameters
    ----------
    p : x,y coords of a point
    line_section : x,y coords of two points defining a line pass point p
    
    Returns
    -------
    ind_pos : -1|0|1, out on side of the first point(-1), within the line
        section(0), out on side of the last point(1)

    """
    p = np.array(point).flatten()
    d_p_p0 = norm(p-line_section[0])
    d_p_p1 = norm(p-line_section[1])
    d_p0_p1 = norm(line_section[0]-line_section[1])
    if np.logical_and(d_p_p0 < d_p_p1, d_p_p1 > d_p0_p1):
        ind_pos = -1 # out on p0 side
    elif np.logical_and(d_p_p1 < d_p_p0, d_p_p0 > d_p0_p1):
        ind_pos = 1  # out on p1 side
    else:
        ind_pos = 0
    return ind_pos





#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Created on Wed May 26 14:30:07 2021
# @author: Xiaodong Ming

"""
remove_block
===========
To do:
    remove overhead buildings on DEM
Input:
    1. DEM file (map unit:m)
    2. shape file marking overhead building with a quadrangle
    
Example

# shape_file = '/Users/ming/Dropbox/Python/Data/overhead_buildings.shp'
# dem_file = '/Users/ming/Dropbox/Python/RiverChannel/CA1_5m.tif'
shape_file = 'file:///Users/ming/OneDrive - Loughborough University/Tyneside/SHP/OverheadBuildings.shp'
dem_file = '/Users/ming/OneDrive - Loughborough University/Tyneside/DSM_DTM_BM_merged.tif'
dem_new, channel_dict_list = remove_overhead_buildings(dem_file, shape_file)

#%%
dem_obj = Raster(dem_file)
shape_polygon = fiona.open(shape_file)

#%% for one feature
one_feature = shape_polygon[0]
line_xy = one_feature['geometry']['coordinates'][0]
line_xy = np.array(line_xy)
bound = get_polygon_bound(line_xy, expand_edge=dem_obj.cellsize*2)
dem_clip = dem_obj.rect_clip(bound)
channel_dict = construct_river_section(dem_clip, line_xy)
if 'adjust_val' in one_feature['properties'].keys():
    adjust_val = one_feature['properties']['adjust_val']
else:
    adjust_val = None
dem_new = get_new_z(dem_clip, channel_dict, adjust_val)

#%% 
fig, ax = dem_obj.hillshade()
for channel_dict in channel_dict_list:
    pts_coords = channel_dict['polygon']
    ax.plot(pts_coords[:, 0], pts_coords[:, 1], 'r') 
# ax.plot(channel_dict['bank0'][:, 0], channel_dict['bank0'][:, 1], '*')
# ax.plot(channel_dict['bank1'][:, 0], channel_dict['bank1'][:, 1], '*') 
# ax.plot(channel_dict['cross0'][:, 0], channel_dict['cross0'][:, 1], 's') 
# ax.plot(channel_dict['cross1'][:, 0], channel_dict['cross1'][:, 1], 's') 
#%%
pts_coords = shape_polygon[0]['geometry']['coordinates'][0]
pts_coords = np.array(pts_coords)
#%%
pts_coords = channel_dict['polygon']
fig, ax = dem_clip.mapshow()
ax.plot(pts_coords[:, 0], pts_coords[:, 1], 'r*')    

"""

#%% load files
import fiona
import copy
import numpy as np
from . import Raster
from numpy.linalg import norm
from .channel_geometry import point2segment, remove_duplicate_rows
from .channel_geometry import distance_p2l, in_channel
from .convert_coords import xy2bc
#%
def remove_overhead_buildings(dem_obj, shape_polygon):
    """ remove overhead buildings on DEM file
    
    Parameters
    ----------
    dem_obj : a Raster object for DEM
    shape_polygon : a Collection object of fiona module 
        or a string for polygon shapefile.
        Each polygon must consists of four points, namely a quadrangle.
        If the polygon has property 'adjust_val', then all the cell inside the
        polygon will be adjusted according to the given value
    
    Returns
    -------
    dem_new : a Raster object with rectified DEM

    """
    if type(dem_obj) is str:
        dem_obj = Raster(dem_obj)
    dem_obj = copy.deepcopy(dem_obj)
    if type(shape_polygon) is str:
        shape_polygon = fiona.open(shape_polygon)
    # rasterize shapefile
    # ind_array = dem_obj.rasterize(shape_polygon)
    channel_dict_list = []
    N = 0
    for one_feature in shape_polygon:
        if one_feature['geometry'] is None:
            continue
        line_xy = one_feature['geometry']['coordinates'][0]
        line_xy = np.array(line_xy)
        bound = get_polygon_bound(line_xy, expand_edge=dem_obj.cellsize*2)
        dem_clip = dem_obj.rect_clip(bound)
        channel_dict = construct_river_section(dem_clip, line_xy)
        if 'adjust_val' in one_feature['properties'].keys():
            adjust_val = one_feature['properties']['adjust_val']
        else:
            adjust_val = None
        dem_new = get_new_z(dem_clip, channel_dict, adjust_val)
        dem_new.paste_on(dem_obj)
        channel_dict_list.append(channel_dict)
        print(N)
        N = N+1
    return dem_obj, channel_dict_list

def get_polygon_bound(polygon_xy, expand_edge=None):
    """ get (left, right, bottom, top) of a polygon feature
    """
    left = polygon_xy[:, 0].min()
    right = polygon_xy[:, 0].max()
    bottom = polygon_xy[:, 1].min()
    top = polygon_xy[:, 1].max()
    if expand_edge is not None:
        bound = (left-expand_edge, right+expand_edge, 
                 bottom-expand_edge, top+expand_edge)
    else:
        bound = (left, right, bottom, top)
    return bound

def get_cells_on_line1(dem_obj, segment):
    """ find cells across a line
    the distance between the cell centre and the land is no larger than half
    cellsize
    dem_obj: a Raster object for DEM
    segment: a 2X2 array providing the coords of two points
    return 

    """
    cells_x, cells_y = dem_obj.to_points()
    cells_xy = np.c_[cells_x.flatten(), cells_y.flatten()]
    # find ending cells of the line section
    d0 = norm(cells_xy - segment[0], axis=1)
    d1 = norm(cells_xy - segment[1], axis=1)
    ind_ends = np.logical_or(d0 <= d0.min(), d1 <= d1.min())
    # exclude cells far from the line section
    d2 = distance_p2l(np.array(segment[0:2]), cells_xy)
    _, ind_pos = point2segment(cells_xy, np.array(segment[0:2]))
    ind_inline = np.logical_and(d2 <= dem_obj.cellsize/2, ind_pos)
    ind_online = np.logical_or(ind_ends, ind_inline)
    cell_xyz = np.c_[cells_xy[ind_online], 
                     dem_obj.array.flatten()[ind_online]]
    # sort points
    d3 = norm(cells_xy[ind_online]-segment[0], axis=1)
    ind = d3.argsort()
    cell_xyz = cell_xyz[ind]

    return cell_xyz

def get_cells_on_line(dem_clip, segment):
    """ find cells across a line
    the distance between the cell centre and the land is no larger than half
    cellsize
    dem_clip: a Raster object for DEM
    segment: a 2X2 array providing the coords of two points
    return 

    """
    one_line = {"type": "LineString", 
                "coordinates": [segment[0], segment[1]]}
    shape_lines = [one_line]
    ind_array = dem_clip.rasterize(shape_lines)
    cells_x, cells_y = dem_clip.to_points()
    cell_xyz = np.c_[cells_x[ind_array], cells_y[ind_array],
                     dem_clip.array[ind_array]]
    d = norm(cell_xyz[:, :2]-segment[0], axis=1)
    ind = d.argsort()
    cell_xyz = cell_xyz[ind]
    return cell_xyz

def construct_river_section(dem_clip, polygon_xy):
    """ construct a river channel based on polygon points (4) 
    
    Parameters
    ----------
    dem_clip : a Raster object for DEM
    polygon_xy : TYPE

    Returns
    -------
    channel_dict : dict with five keys, showing the channel line points 

    """
    polygon_xy = np.array(polygon_xy)
    polygon_xy = remove_duplicate_rows(polygon_xy)
    if polygon_xy.shape[0] > 4: # if the polygon is not quadrangle
        channel_dict = {'polygon':polygon_xy}
    else:
        cl0 = polygon_xy[[0, 1]]
        cl0_xyz = get_cells_on_line(dem_clip, cl0)
        cl1 = polygon_xy[[3, 2]]
        cl1_xyz = get_cells_on_line(dem_clip, cl1)
        bl0 = polygon_xy[[0, 3]]
        bl0_xyz = get_cells_on_line(dem_clip, bl0)
        bl1 = polygon_xy[[1, 2]]
        bl1_xyz = get_cells_on_line(dem_clip, bl1)
        # identify crossline and bankline according to the median z value
        # crossline should have a lower median z value
        z_c = np.concatenate([cl0_xyz[:,2], cl1_xyz[:,2]])
        z_b = np.concatenate([bl0_xyz[:,2], bl1_xyz[:,2]])
        if np.median(z_c) < np.median(z_b):
            bline0 = bl0_xyz[:, :2]
            bline1 = bl1_xyz[:, :2]
            cline0 = cl0_xyz
            cline1 = cl1_xyz
        else:
            bline0 = cl0_xyz[:, :2]
            bline1 = cl1_xyz[:, :2]
            cline0 = bl0_xyz
            cline1 = bl1_xyz
        # construct river section    
        all_points = np.r_[bline0, cline1[:, :2], np.flipud(bline1),
                               np.flipud(cline0[:, :2])]
        channel_dict = {'bank0':bline0, 'bank1':bline1,
                        'cross0':cline0, 'cross1':cline1,
                        'polygon':all_points}
    return channel_dict

def get_new_z(dem_obj, channel_dict, adjust_val):
    cells_x, cells_y = dem_obj.to_points()
    cells_xy = np.c_[cells_x.flatten(), 
                     cells_y.flatten()]
    cells_z = dem_obj.array.flatten()
    ind = in_channel(channel_dict, cells_xy, radius=0.00001)
    xy_cells_vec = cells_xy[ind]
    z_cells_vec = cells_z[ind]
    if adjust_val is None:
        bs, cs, zs = xy2bc(xy_cells_vec, channel_dict, 
                           max_error=dem_obj.cellsize*2)
        z_cells_vec[~np.isnan(zs)] = zs[~np.isnan(zs)]
    else:
        z_cells_vec = z_cells_vec+adjust_val
    cells_z[ind] = z_cells_vec
    dem_obj.array = np.reshape(cells_z, dem_obj.array.shape)
    return dem_obj

def main():
    print('Package to show grid data')

if __name__=='__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Created on Mon Dec  7 12:55:34 2020
# @author: Xiaodong Ming

"""
tutorial
===========
To do:
    [objective of this script]

"""
#%%
import pickle
import numpy as np
import matplotlib.pyplot as plt
from convert_coords import cross_section2grid_elevation
from demo_functions import draw_arrow

# load and show input bankline and cross-section lines
with open('sample/sample_data.pkl', 'rb') as f:
    sample_data = pickle.load(f)
bankline0 = sample_data['bankline0']
bankline1 = sample_data['bankline1']
crossline_list = sample_data['crosslines']
#%% Composite function to convert cross lines to DEM grid cells
z_grid, header = cross_section2grid_elevation(bankline0, bankline1, 
                                              crossline_list, 
                                              cellsize=0.2, max_error=0.1)
#%% show results with bank lines and cross lines
import hipims_io as hp
dem_obj = hp.Raster(array=z_grid, header=header)
fig, ax = dem_obj.hillshade()
ax.plot(bankline0[:, 0], bankline0[:, 1], 'r')
ax.plot(bankline1[:, 0], bankline1[:, 1], 'r')
for n in np.arange(len(crossline_list)):
    line_data = crossline_list[n]
    ax.plot(line_data[:, 0], line_data[:, 1], marker='.')

#%% Step-by-step tutorial
import convert_coords as cc
import channel_geometry as cg
#%% show banklines and cross section lines with directions of their vertex
fig1, axs = plt.subplots(2, 1)
ax1 = axs[0]
ax2 = axs[1]
ax1.plot(bankline0[:, 0], bankline0[:, 1], 'r')
draw_arrow(ax1, bankline0, width=0.2)
ax1.plot(bankline1[:, 0], bankline1[:, 1], 'r')
draw_arrow(ax1, bankline1, width=0.2)
for n in np.arange(len(crossline_list)):
    line_data = crossline_list[n]
    ax1.plot(line_data[:, 0], line_data[:, 1], marker='.')
    draw_arrow(ax1, line_data, width=0.2)
    ax1.annotate(str(n), xy=line_data[0][:2])
    xy1d = cc.normalise_xy_coords(line_data[:, 0], line_data[:, 1])
    ax2.plot(xy1d, line_data[:, 2])
ax1.set_aspect('equal', adjustable='box')

#%% Sort bankline points to the same direction, trim and sort cross lines
bl0, bl1, cl_list = cg.preprocessing_data(bankline0, bankline1, crossline_list)
fig1, ax1 = plt.subplots(1)
ax1.plot(bl0[:, 0], bl0[:, 1], 'r')
draw_arrow(ax1, bl0, width=0.2)
ax1.plot(bl1[:, 0], bl1[:, 1], 'r')
draw_arrow(ax1, bl1, width=0.2)
for n in np.arange(len(cl_list)):
    line_data = cl_list[n]
    if line_data.size>0:
        ax1.plot(line_data[:, 0], line_data[:, 1], marker='.')
        draw_arrow(ax1, line_data, width=0.2)
        ax1.annotate(str(n), xy=line_data[0][:2])
ax1.set_aspect('equal', adjustable='box')

#%% split river polygon using crosslines
channel_list = cc.split_channel(cl_list, bl0, bl1)
fig, ax = plt.subplots(1)
n=0
for channel_dict in channel_list:
    ax.plot(channel_dict['bank0'][:, 0], channel_dict['bank0'][:, 1], 
            marker='o')
    draw_arrow(ax, channel_dict['bank0'], width=0.2)
    ax.plot(channel_dict['bank1'][:, 0], channel_dict['bank1'][:, 1], 
            marker='o')
    draw_arrow(ax, channel_dict['bank1'], width=0.2)
    ax.plot(channel_dict['cross0'][:, 0], channel_dict['cross0'][:, 1], 
            marker='.')
    draw_arrow(ax, channel_dict['cross0'], width=0.2)
    ax.annotate(str(n), xy=channel_dict['cross0'][0][:2])
    n=n+1
    ax.plot(channel_dict['cross1'][:, 0], channel_dict['cross1'][:, 1], 
            marker='.')
    draw_arrow(ax, channel_dict['cross1'], width=0.2)
ax.set_aspect('equal', adjustable='box')
#%% for each river channel section
max_error = 0.2
n = 0
channel_dict = channel_list[n]
bline0 = channel_dict['bank0']
bline1 = channel_dict['bank1']
# break bank line
_, break_points_0 = cg.break_bankline(bline0, max_error)
_, break_points_1 = cg.break_bankline(bline1, max_error)
fig, ax = plt.subplots(1)
ax.plot(bline0[:, 0], bline0[:, 1])
ax.plot(break_points_0[:, 0], break_points_0[:, 1], marker='o')
ax.plot(bline1[:, 0], bline1[:, 1])
ax.plot(break_points_1[:, 0], break_points_1[:, 1], marker='s')
#%% project the bank break points to the opposite bank
break_points_exp_0 = cc.expand_break_points(break_points_0,
                                             break_points_1[1:-1])
break_points_exp_1 = cc.expand_break_points(break_points_1,
                                             break_points_0[1:-1])
fig, ax = plt.subplots(1)

ax.plot(break_points_exp_0[:, 0], break_points_exp_0[:, 1], marker='o',
           mfc='none')
ax.scatter(break_points_0[:, 0], break_points_0[:, 1], marker='*')
ax.plot(break_points_exp_1[:, 0], break_points_exp_1[:, 1], marker='s',
           mfc='none')
ax.scatter(break_points_1[:, 0], break_points_1[:, 1], marker='*')

#%% discretize river channel to polygons
river_polys, breaks_0, breaks_1 = cc.discretize_river_channel(bline0, 
                                            bline1, max_error)
for geom in river_polys.geoms:
        xs, ys = geom.exterior.xy
        plt.fill(xs, ys, alpha=0.5)
plt.gca().set_aspect('equal', adjustable='box')
#%% create a river topography grid
grid_header = sample_data['header']
cellsize = grid_header['cellsize']
nrows = grid_header['nrows']
ncols = grid_header['ncols']
x00 = grid_header['xllcorner']+cellsize/2
y00 = grid_header['yllcorner']+nrows*cellsize-cellsize/2
x_vect = np.linspace(x00, x00+cellsize*ncols, num=ncols, endpoint=False)
y_vect = np.linspace(y00, y00-cellsize*nrows, num=nrows, endpoint=False)
x_grid, y_grid = np.meshgrid(x_vect, y_vect)
#%% show relative coordinates of grid cells
xy_cells_vec = np.c_[x_grid.flatten(), y_grid.flatten()]
bc_pos = cc.convert_coords_2d(xy_cells_vec, bline0, bline1, max_error)
fig, ax = plt.subplots(1)
points_b = bc_pos[:,0]
points_c = bc_pos[:,1]
points_c[points_c>0.5] = 1-points_c[points_c>0.5]
points_z = np.round(points_c*10)
z_grid = points_z.reshape(x_grid.shape)
ax.contourf(x_grid, y_grid, z_grid, 10, cmap='YlGnBu', alpha=1)
points_z = np.round(points_b*20)
z_grid = points_z.reshape(x_grid.shape)
ax.contourf(x_grid, y_grid, z_grid, 20, cmap='Reds', alpha=0.5)
ax.set_aspect('equal', adjustable='box')

#%%
bc_pos = cc.convert_coords_2d(xy_cells_vec, bline0, bline1, max_error)
bs = bc_pos[:, 0]
grid_b = bs.reshape(x_grid.shape)
cs = bc_pos[:, 1]
grid_c = cs.reshape(x_grid.shape)
cline0 = channel_dict['cross0']
cline1 = channel_dict['cross1']
zs = cc.bc_interp_2d(bs, cs, cline0, cline1)
z_grid = zs.reshape(x_grid.shape)

#%
# from mpl_toolkits.mplot3d import Axes3D
bline0_z = np.linspace(cline0[0, 2], cline1[0, 2], num=bline0.shape[0], 
                       endpoint=False)
bline1_z = np.linspace(cline0[-1, 2], cline1[-1, 2], num=bline1.shape[0], 
                       endpoint=False)
fig = plt.figure()
ax = fig.add_subplot(111, projection='3d')

ax.plot(bline0[:, 0], bline0[:, 1], bline0_z, 'r')
ax.plot(bline1[:, 0], bline1[:, 1], bline1_z, 'b')
ax.plot(cline0[:, 0], cline0[:, 1], cline0[:, 2], '-.')
ax.plot(cline1[:, 0], cline1[:, 1], cline1[:, 2], '--')
xs = xy_cells_vec[:, 0]
ys = xy_cells_vec[:, 1]
ax.scatter(xs, ys, zs, '.', s=1)
ax.plot_trisurf(xs, ys, zs, linewidth=0.2, color=(0,0,0,0), edgecolor='Grey')
ax.view_init(44, -57)
plt.tight_layout()
# fig.savefig('interpolated_trisurf', dpi=300)
#%%
header = sample_data['header']
array = cc.cross_section2grid_elevation(bankline0, bankline1, crossline_list, 
                                          header, max_error=0.2)

from hipims_io import Raster
dem = Raster(array=array, header=header)
dem.mapshow()
#%%
fig = plt.figure()
ax = fig.add_subplot(111, projection='3d')
z_grid = dem.array
ax.scatter(x_grid, y_grid, z_grid, '.', s=1)
ax.plot_trisurf(x_grid.flatten(), y_grid.flatten(), z_grid.flatten(),
                linewidth=0.2, color=(0,0,0,0), edgecolor='Grey')
for crossline in crossline_list:
    ax.plot(crossline[:, 0], crossline[:, 1], crossline[:, 2], '*')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Created on Thu Mar 18 16:14:45 2021
# @author: Xiaodong Ming

"""
Tutorial_carlisle
===========
To do:
    use the Eden river cross sections in Carlisle to generate bed terrain

"""
import numpy as np
import scipy.io as sio
import matplotlib.pyplot as plt
from hydro_raster.convert_coords import cross_section2grid_elevation
from hydro_raster.Raster import Raster
#% load mat data
mat_fname = 'CarlisleRiverChannelData.mat'
mat_contents = sio.loadmat(mat_fname)
channelBoundaryLines = mat_contents['channelBoundaryLine']
crossSectionLines = mat_contents['crossSectionLine']
channelBoundary = mat_contents['channelBoundary']
#% Get bankline points
points_x = channelBoundaryLines[0]['X'][0]
points_y = channelBoundaryLines[0]['Y'][0]
bankline0 = np.c_[points_x.flatten(), points_y.flatten()]
bankline0 = bankline0[~np.isnan(bankline0[:, 0])]

points_x = channelBoundaryLines[1]['X'][0]
points_y = channelBoundaryLines[1]['Y'][0]
bankline1 = np.c_[points_x.flatten(), points_y.flatten()]
bankline1 = bankline1[~np.isnan(bankline1[:, 0])]

#% Get crossline points
crossline_list = [a[0] for a in crossSectionLines]

#%% show banklines and cross section lines with directions of their vertex
fig, ax = plt.subplots(1)
ax.plot(bankline0[:, 0], bankline0[:, 1], 'r')
ax.plot(bankline1[:, 0], bankline1[:, 1], 'r')
for n in np.arange(len(crossline_list)):
    line_data = crossline_list[n]
    ax.plot(line_data[:, 0], line_data[:, 1], marker='.')
    ax.annotate(str(n), xy=line_data[0][:2])
ax.set_aspect('equal', adjustable='box')
#%% create terrain data from cross section lines and bank lines
array, header = cross_section2grid_elevation(bankline0, bankline1, 
                                             crossline_list, cellsize=2,
                                             max_error=0.1)
#%% show interpolated river topography with source lines
obj_dem = Raster(array=array, header=header)
fig, ax = obj_dem.mapshow()
ax.plot(bankline0[:, 0], bankline0[:, 1], 'r')
ax.plot(bankline1[:, 0], bankline1[:, 1], 'r')
for n in np.arange(len(crossline_list)):
    line_data = crossline_list[n]
    # ax.plot(line_data[:, 0], line_data[:, 1], 'k-')
#%% show preprocessed data: sorted and trimmed lines
from demo import draw_arrow
from channel_geometry import preprocessing_data
bl0, bl1, cl_sort = preprocessing_data(bankline0, bankline1, crossline_list)
fig, ax = plt.subplots(1)
ax.plot(bl0[:, 0], bl0[:, 1], 'r')
draw_arrow(ax, bankline0, width=0.2)
ax.plot(bl1[:, 0], bl1[:, 1], 'r')
draw_arrow(ax, bankline1, width=0.2)
for n in np.arange(len(cl_sort)):
    line_data = cl_sort[n]
    ax.plot(line_data[:, 0], line_data[:, 1], 'b:')
    draw_arrow(ax, line_data, width=0.2)
    ax.annotate(str(n), xy=line_data[0][:2])
ax.set_aspect('equal', adjustable='box')

#%% Show splitted river channels
from channel_geometry import split_channel
from channel_geometry import break_bankline, discretize_river_section
from channel_geometry import match_break_points_by_relative_pos
# from channel_geometry import match_break_points
river_sections = split_channel(cl_sort, bl0, bl1)
max_error = 2
fig, ax = plt.subplots(1)
for n in [19, 20, 21, 22]: #np.arange(len(river_sections)): 
    bl0_sec = river_sections[n]['bank0']
    bl1_sec = river_sections[n]['bank1']
    cl0 = river_sections[n]['cross0']
    cl1 = river_sections[n]['cross1']
    _, bl0_broken = break_bankline(bl0_sec, max_error)
    _, bl1_broken = break_bankline(bl1_sec, max_error)
    bl0_broken_e, bl1_broken_e = match_break_points_by_relative_pos(bl0_broken, 
                                                                    bl1_broken)
    # bl0_broken_e, bl1_broken_e = match_break_points(bl0_broken, bl1_broken)
    river_polys = discretize_river_section(bl0_broken_e, bl1_broken_e)
    ax.plot(bl0_broken[:, 0], bl0_broken[:, 1], '--',
                marker='o', mfc='None') #, c='b'
    ax.plot(bl1_broken[:, 0], bl1_broken[:, 1], '--', 
                marker='o', mfc='None') #, c='b'
    ax.plot(bl0_broken_e[:, 0], bl0_broken_e[:, 1], '--',
                marker='*', mfc='None') #, c='b'
    ax.plot(bl1_broken_e[:, 0], bl1_broken_e[:, 1], '--', 
                marker='*', mfc='None') #, c='b'
    ax.plot(cl0[:, 0], cl0[:, 1], 'k-', linewidth=0.5) #, c='b'
    ax.plot(cl1[:, 0], cl1[:, 1], 'k-', linewidth=0.5) #, c='b'
    
    # xy_anno = np.median(river_sections[n]['polygon'], axis=0)
    # ax.annotate(str(n), xy=xy_anno)
    # for i in np.arange(bl1_broken.shape[0]):
    #     xy_p = bl1_broken[i]
    #     ax.annotate(str(i), xy=xy_p)   
    #%
    for geom in river_polys.geoms:
        xs, ys = geom.exterior.xy
        ax.fill(xs, ys, alpha=0.5)
ax.set_aspect('equal', adjustable='box')
fname = 'split_channel_err'+str(max_error)+'m_'+'.png'
# fig.savefig(fname, bbox_inches='tight', pad_inches=0, dpi=2000)
#%% Show relative bc coordinates of river points
from channel_geometry import get_bounding_box
from convert_coords import interp_one_reach, generate_cells_xy
fig, ax = plt.subplots(1)

cellsize = 5
for n in [17, 18, 19, 20, 21, 22, 23, 24, 25]:
    one_reach = river_sections[n]
    bounding_box = get_bounding_box(one_reach['polygon'])
    x_grid, y_grid, header = generate_cells_xy(bounding_box, cellsize)
    bs, cs, zs = interp_one_reach(one_reach, x_grid, y_grid, 
                                  max_error=0.1)
    ax.plot(one_reach['polygon'][:, 0], one_reach['polygon'][:, 1], 'k-')
    ax.contour(x_grid, y_grid, bs, 10, cmap='winter')
    ax.contour(x_grid, y_grid, cs, 10, cmap='autumn')
ax.set_aspect('equal', adjustable='box')


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Created on Tue Mar 23 19:48:38 2021
# @author: Xiaodong Ming

"""
channel_geometry
===========
To do:
    deal with polyline and points, sort bank lines, cross-section lines
    construct closed polygon with bank lines and cross lines
    identify geometry relations
"""
import numpy as np
from matplotlib import path
from numpy.linalg import norm
from shapely import geometry
from .coords_interp import line_1d_2d
def preprocessing_data(bankline0, bankline1, crossline_list):
    """ Preprocess raw data
    0.  Remove vertex without xy coordiantes (nans in points-array)
    1.	Unify the direction of two bank lines
    2.	Identify the first and the last cross-section lines, referring to the 
            bank line direction
    3.	Trime bank lines in their two ends
    4.  Cut outside vertices of cross-section lines and unify their directions
    5.	Sort the list of cross-section lines towards the bank line direction

    Parameters
    ----------
    bankline0 : 2-col array, xy coords of bank line vertex
    bankline1 : 2-col array, xy coords of bank line vertex
    crossline_list : list of 2-col arrays, xy coords of cross line vertex

    Returns
    -------
    bl0, bl1, cl_list_sort: sorted and trimmed bank lines and cross lines

    """
    # 0 Remove nan value vertex
    bl0 = bankline0[~np.isnan(bankline0[:, -1]), :]
    bl1 = bankline1[~np.isnan(bankline1[:, -1]), :]
    cl_list = []
    for crossline in crossline_list:
        crossline = crossline[~np.isnan(crossline[:, -1]), :]
        cl_list.append(crossline)
    # 1 Sort banklien vertex towards the same stream direction
    bl1 = sort_line_points(bl0[0], bl1) # first point of bl0 is the reference
    # 2 Find the first and last cross line
    ind_cl0, ind_cl1 = identify_crossline_ends(bl0, bl1, cl_list)
    # 3 Trime bank lines in their two ends
    bl0 = trim_bankline(bl0, cl_list[ind_cl0], cl_list[ind_cl1])
    bl1 = trim_bankline(bl1, cl_list[ind_cl0], cl_list[ind_cl1])
    # 4 Trime and sort cross line
    river_channel, cl0_cut, cl1_cut = construct_channel_polygon(bl0, bl1, 
                                            cl_list[ind_cl0], cl_list[ind_cl1])
    for ele in sorted([ind_cl0, ind_cl1], reverse = True):
        del cl_list[ele]
    cl_list_cut = trim_crossline_points(river_channel, cl_list)
    cl_list_cut.insert(0, cl0_cut)
    cl_list_cut.insert(-1, cl1_cut)
    # 5 Sort the list of cross-section lines towards the bank line directio
    cl_list_sort = sort_crossline_list(cl_list_cut, bl0)

    return bl0, bl1, cl_list_sort

def sort_line_points(point_refer, bankline):
    """Sort (bank) line points refer to the original point
    point_refer: 2-element array, the origin reference point of bank lines
    """
    bankline = bankline[~np.isnan(bankline[:,0]), :]
    p0 = bankline[0]
    p1 = bankline[-1]
    if norm(point_refer-p0) > norm(point_refer-p1):
        bankline_sorted = np.flipud(bankline)
    else:
        bankline_sorted = bankline+0
    return bankline_sorted

def identify_crossline_ends(bl0, bl1, crossline_list):
    """Identify the first and the last cross-section lines, referring to the 
            bank line direction
    Parameters
    ----------
    bl0 and bl1 are bankline vertices sorted towards the same direction
    crossline_list: contains middle point of each crossline
    Returns
    -------
    ind_cl0, ind_cl1: index of the first / last crossline

    """
    cl_mid_list = [] # middle point of each cross line
    for crossline in crossline_list:
        cl_mid = crossline[[0, -1], :2]
        cl_mid = cl_mid.mean(axis=0)
        cl_mid_list.append(cl_mid)
    cl_mid = np.array(cl_mid_list)
    point_ref = (bl0[0]+bl1[0])/2
    distance = norm(cl_mid-point_ref, axis=1)
    ind = distance.argsort()
    ind_cl0 = ind[0] # index of the first cross line
    point_ref = (bl0[-1]+bl1[-1])/2
    distance = norm(cl_mid-point_ref, axis=1)
    ind = distance.argsort()
    ind_cl1 = ind[0] # index of the alst cross line
    return ind_cl0, ind_cl1

def trim_bankline(bline, cline0, cline1):
    """make the ends of bankline be connected with the crosslines in two ends 
    bline: bankline
    cline0: crossline on the beginning
    cline1: crossline on the ending
    """
    # the left end
    straight_line = cline0[[0, -1], :2]
    inter_point, inter_ind = find_inter_point_on_polyline(bline, straight_line,
                                                          True)
    if inter_ind is None:
        b_segment = bline[0:2]
        inter_point = get_intersection(b_segment, straight_line)
        c_ind = nearest_points(inter_point, cline0[:, :2])
        bline = np.r_[cline0[c_ind, :2].reshape(1,2), bline]
    else:
        bline = bline[inter_ind:]
        bline[0] = inter_point    
    
    # the right end
    straight_line = cline1[[0, -1], :2]
    inter_point, inter_ind = find_inter_point_on_polyline(bline, straight_line,
                                                          True)
    if inter_ind is None:
        b_segment = bline[-2:]
        inter_point = get_intersection(b_segment, straight_line)
        c_ind = nearest_points(inter_point, cline1[:, :2])
        bline = np.r_[bline, cline1[c_ind, :2].reshape(1,2)]
    else:
        bline = bline[:inter_ind+1]
        bline[-1] = inter_point
    indexes = np.unique(bline, return_index=True, axis=0)[1]
    bline = [bline[i] for i in sorted(indexes)]# remove duplicated points
    bline = np.array(bline)
    return bline

def construct_channel_polygon(bl0, bl1, cl0, cl1):
    """ Cut two cross lines and create a Path object for river channel
    bl0, bl1 are sorted and trimmed banklines, cl0 and cl1 are the first and
    the last crossline without cutting
        point10--------------bank1-------------point11
           |                                      |
           |                                      |
        cross0                                  cross1
           |                                      |
           |                                      |
        point00--------------bank0-------------point01
    """
    ind_c00 = nearest_points(bl0[0], cl0[:, :2])
    ind_c10 = nearest_points(bl1[0], cl0[:, :2])
    ind_c01 = nearest_points(bl0[-1], cl1[:, :2])
    ind_c11 = nearest_points(bl1[-1], cl1[:, :2])
    if ind_c00 > ind_c10:
        cl0_cut = cl0[ind_c10:ind_c00+1]
        cl0_cut = np.flipud(cl0_cut)
    else:
        cl0_cut = cl0[ind_c00:ind_c10+1]
    if ind_c01 > ind_c11:
        cl1_cut = cl1[ind_c11:ind_c01+1]
        cl1_cut = np.flipud(cl1_cut)
    else:
        cl1_cut = cl1[ind_c01:ind_c11+1]
    cl0_cut[0, :2] = bl0[0]
    cl0_cut[-1, :2] = bl1[0]
    cl1_cut[0, :2] = bl0[-1]
    cl1_cut[-1, :2] = bl1[-1]
    xy_points = np.r_[bl0, cl1_cut[:, :2], 
                      np.flipud(bl1), np.flipud(cl0_cut[:, :2])]
    channel_points = path.Path(xy_points)
    return channel_points, cl0_cut, cl1_cut

def trim_crossline_points(river_channel, crossline_list):
    cl_list_cut = []
    N=0
    for crossline in crossline_list:
        points_xy = crossline[:, :2]
        ind = river_channel.contains_points(points_xy, radius=-0.001)
        crossline = crossline[ind, :]
        if ind.sum()==0:
            print(N)
            raise ValueError('crossline fully cut, check the radius value')
        N = N+1
        # remove repeat
        crossline = remove_duplicate_rows(crossline)
        cl_list_cut.append(crossline)
    return cl_list_cut
    
def sort_crossline_list(cl_list_cut, bline):
    
    # find two central points of cross line
    ind_b_list = []
    cl_mid_list = []
    for crossline in cl_list_cut:
        cl_mid = np.mean(crossline[:, :2], axis=0)
        ind_2 = nearest_points(cl_mid, crossline[:, :2], 2)
        cl_segm = crossline[ind_2, :2]
        _, ind_b = find_inter_point_on_polyline(bline, cl_segm)
        if ind_b is None:
            ind_b = nearest_points(cl_mid, bline, 1)
        ind_b_list.append(ind_b)
        cl_mid_list.append(cl_mid)
    # check duplicated ind
    cl_mid_list = np.array(cl_mid_list)
    ind_b_list = np.array(ind_b_list).astype('float64')
    u, c = np.unique(np.array(ind_b_list), return_counts=True)
    dup_ind = u[c>1] # duplicated ind
    if dup_ind.size>0:
        for ind_b_dup in dup_ind:
            ind_b_dup = int(ind_b_dup)
            cl_mid_dup = cl_mid_list[ind_b_list == ind_b_dup]
            b_point = bline[ind_b_dup].reshape(1,2)
            b_point = np.repeat(b_point, cl_mid_dup.shape[1], axis=0)
            d_cl2bl =  norm(cl_mid_dup-b_point, axis=1)
            add_v = d_cl2bl.argsort()
            add_v = add_v*(0.5/add_v.max())
            ind_b_list[ind_b_list == ind_b_dup] = ind_b_dup+add_v
            
    ind_sort = np.array(ind_b_list).argsort()
    cl_sort = [cl_list_cut[ind] for ind in ind_sort]
    return cl_sort

def split_channel(cl_list, bl0, bl1):
    """ Split river channel to river sections by cross-section lines
    and eidt end points of each line to make them connected with each other
    all input lines must be trimmed and sorted before
    Return:
        channel_list Dict with five keys:
        channel_dict = {'bank0':bline0, 'bank1':bline1,
                        'cross0':cline0, 'cross1':cline1,
                        'polygon':all_points
        point10--------------bank1-------------point11
            |                                      |
            |                                      |
        cross0                                  cross1
            |                                      |
            |                                      |
        point00--------------bank0-------------point01
    """
    bl0 = bl0+0
    bl1 = bl1+0
    channel_list = []
    for n in np.arange(len(cl_list)-1):
        cline0 = cl_list[n]
        cl_0 = cline0[[0, -1], :2]
        cline1 = cl_list[n+1]
        cl_1 = cline1[[0, -1], :2]
        p00, ind00 = find_inter_point_on_polyline(bl0, cl_0, avoid_none=True)
        p10, ind10 = find_inter_point_on_polyline(bl1, cl_0, avoid_none=True)
        p11, ind11 = find_inter_point_on_polyline(bl1, cl_1, avoid_none=True)
        p01, ind01 = find_inter_point_on_polyline(bl0, cl_1, avoid_none=True)
        # change the xy cooords of crossline endpoints to intersections
        cline0[0][:2]  = p00
        cline0[-1][:2] = p10
        cline0 = line_flatten(cline0)
        cline1[0][:2]  = p01
        cline1[-1][:2] = p11
        cline1 = line_flatten(cline1)
        bline0 = bl0[ind00:ind01+1]
        if bline0.shape[0]>1:
            if within_segment(p00, bline0[0:2]):
                bline0[0] = p00
            else:
                bline0 = np.r_[p00, bline0]
            if within_segment(p01, bline0[-2:]):
                bline0[-1] = p01
            else:
                bline0 = np.r_[bline0, p01]
        else:
            bline0 = np.r_[p00, bline0, p01]
        bline0 = remove_duplicate_rows(bline0)
        
        bline1 = bl1[ind10:ind11+1]
        if bline1.shape[0] > 1:
            if within_segment(p10, bline1[0:2]):
                bline1[0] = p10
            else:
                bline1 = np.r_[p10, bline1]
            if within_segment(p11, bline1[-2:]):
                bline1[-1] = p11
            else:
                bline1 = np.r_[bline1, p11]
        else:
            bline1 = np.r_[p10, bline1, p11]
        bline1 = remove_duplicate_rows(bline1)
        
        all_points = np.r_[bline0, cline1[:, :2], np.flipud(bline1),
                           np.flipud(cline0[:, :2])]
        channel_dict = {'bank0':bline0, 'bank1':bline1,
                        'cross0':cline0, 'cross1':cline1,
                        'polygon':all_points}
        channel_list.append(channel_dict)
    return channel_list

def line_flatten(polyline):
    # flatten polyline  as a straignt line refer to its two ends
    line_obj = line_1d_2d(polyline)
    if norm(polyline[0, :2] - polyline[-1, :2])==0:
        line_obj0 = line_1d_2d(polyline[[0, -2], :2])
    else:
        line_obj0 = line_1d_2d(polyline[[0, -1], :2])
    xy = line_obj0.d_to_xy(line_obj.d)
    polyline_new = line_obj.xy+0
    polyline_new[:, :2] = xy
    return polyline_new
    
#%% discretise river section
def break_bankline(bankline, max_error):
    """discretise bankline to a orderd line segments
    bankline: X and Y coordinates of bank line points
    max_error: the max distance between line points to straight bank line
    Return:
        break_ind: index of the break points
        broken_line: 2-col array
    """
    break_ind = []
    ind_0 = 0 # start index of a bank section
    ind_1 = bankline.shape[0] # end index of a bank section
    while ind_0 < ind_1:
        section_line = np.array([bankline[ind_0], bankline[ind_1-1]])
        points_xy = bankline[ind_0:ind_1]
        distance = distance_p2l(section_line, points_xy)
        if distance.max() > max_error:
            ind_1 = distance.argmax()+ind_0
            ind_1 = ind_1.astype('int64')
        else:
            break_ind.append(ind_1)
            ind_0 = ind_1
            ind_1 = bankline.shape[0]
    break_ind = np.r_[0, np.array(break_ind)-1]
    broken_line = bankline[break_ind]
    return break_ind, broken_line

def discretize_river_section(bl0_broken_e, bl1_broken_e):
    """discretize river channel to 3/4-side polygons
    river_polys = discretize_river_section(bl0_broken, bl1_broken)
    Return:
        river_polys: shapely multi-polygon
        break_points_full_b: full break points on line_bottom
        break_points_full_t: full break points on line_top
    """
    # expand break points with projections from break points on the other bank
    # break_points_full_b and break_points_full_t should have the same size
    # bl0_broken_expand, bl1_broken_expand = match_break_points_by_relative_pos(bl0_broken, 
    #                                                           bl1_broken)
    polygon_list = []
    for n in np.arange(1, bl0_broken_e.shape[0]):
        point00 = bl0_broken_e[n-1]
        point01 = bl1_broken_e[n-1]
        point11 = bl1_broken_e[n]
        point10 = bl0_broken_e[n]
        polygon = geometry.Polygon([point00, point01, point11, point10])
        polygon_list.append(polygon)
    section_polys = geometry.MultiPolygon(polygon_list)
    return section_polys

def match_break_points(bl0_broken, bl1_broken):
    """ Expand bank break points with points projected from the other bank
    break_points: to define break points of bank line sections, 
        sequence mattters
    points_to_project: coordinates of points on the other bank
    Return:
        break_points_expand
    """
    bl0_broken = remove_duplicate_rows(bl0_broken)
    bl1_broken = remove_duplicate_rows(bl1_broken)
    bl0_broken_e = bl0_broken+0
    bl1_broken_e = bl1_broken+0
    
    pts_repeat = []
    for xy in bl0_broken[1:-1]: # project bl0_broken to bankline1
        point_prj, ind_insert = project_point2polyline(bl1_broken_e, xy)
        if norm(point_prj-bl1_broken_e, axis=1).min()==0:
            pts_repeat.append(point_prj)
        else:
            bl1_broken_e = np.insert(bl1_broken_e, ind_insert, 
                                     point_prj.flatten(), axis=0)
    for pt_repeat in pts_repeat:
        ind_insert = np.where(norm(pt_repeat-bl1_broken_e, axis=1)==0)
        ind_insert = ind_insert[0][0]
        bl1_broken_e = np.insert(bl1_broken_e, ind_insert, 
                                     pt_repeat.flatten(), axis=0)
    
    pts_repeat = []
    for xy in bl1_broken[1:-1]: # project bl1_broken to bankline0
        point_prj, ind_insert = project_point2polyline(bl0_broken_e, xy)
        if norm(point_prj-bl0_broken_e, axis=1).min()==0:
            pts_repeat.append(point_prj)
        else:
            bl0_broken_e = np.insert(bl0_broken_e, ind_insert, 
                                     point_prj.flatten(), axis=0)
    for pt_repeat in pts_repeat:
        ind_insert = np.where(norm(pt_repeat-bl0_broken_e, axis=1)==0)
        ind_insert = ind_insert[0][0]
        bl0_broken_e = np.insert(bl0_broken_e, ind_insert, 
                                     pt_repeat.flatten(), axis=0)
    
    return bl0_broken_e, bl1_broken_e

def match_break_points_by_relative_pos(bl0_broken, bl1_broken):

    obj_0 = line_1d_2d(bl0_broken)
    obj_1 = line_1d_2d(bl1_broken)
    
    bl0_add = obj_0.d_to_xy(obj_1.d[1:-1])
    bl1_add = obj_1.d_to_xy(obj_0.d[1:-1])
    if bl0_add.size >= 2:
        bl0_e = np.r_[obj_0.xy, bl0_add]
        d0_all = np.concatenate([obj_0.d, obj_1.d[1:-1]])
        ind0 = d0_all.argsort()
        bl0_e = bl0_e[ind0]
    else:
        bl0_e = obj_0.xy
    
    if bl1_add.size >= 2:
        bl1_e = np.r_[obj_1.xy, bl1_add]
        d1_all = np.concatenate([obj_1.d, obj_0.d[1:-1]])
        ind1 = d1_all.argsort()
        bl1_e = bl1_e[ind1]
    else:
        bl1_e = obj_1.xy
    return bl0_e, bl1_e
    
def project_point2polyline(polyline, point_xy):
    """
    Project a point to the two nearest line sections and return index

    """
    num_p = polyline.shape[0]
    if num_p == 2: # only one segment
        ind_insert = 1
        point_prj, pos = point2segment(point_xy, polyline)
        if pos:
            point_projected = point_prj.reshape(1, 2)
        else:
            dis_pp = norm(polyline - point_xy, axis=1)
            if dis_pp[0] < dis_pp[1]:
                point_projected = polyline[0].reshape(1, 2)
            else:
                point_projected = polyline[1].reshape(1, 2)
    elif num_p > 2: # two segments
        if num_p == 3:
            ind_nearest = 1
        else:
            ind_nearest2 = nearest_points(point_xy, polyline, 2)
            ind_nearest = ind_nearest2[0]
            if ind_nearest == 0 or ind_nearest == num_p-1:
                ind_nearest = ind_nearest2[1]
            if ind_nearest >= num_p-1:
                ind_nearest = num_p-2
        segment = polyline[ind_nearest-1:ind_nearest+1]
        point_prj_0, pos0 = point2segment(point_xy, segment)
        segment = polyline[ind_nearest:ind_nearest+2]
        point_prj_1, pos1 = point2segment(point_xy, segment)
        # select the best projected point
        if not pos0: # outside
            if not pos1: # outside both two sections, select the nearest point
                # from the three points
                polyline_3pts= polyline[ind_nearest-1:ind_nearest+2]
                ind0 = nearest_points(point_xy, polyline_3pts)
                ind_insert = ind_nearest+ind0
                point_projected = polyline_3pts[ind0]
            else: # project on the second section
                point_projected = point_prj_1
                ind_insert = ind_nearest+1
        else: # onside
            if not pos1: # project on the first section
                point_projected = point_prj_0
                ind_insert = ind_nearest
            else: # onside both two sections, select the nearest projected point
                polyline_3pts= polyline[ind_nearest-1:ind_nearest+2]
                ind0 = nearest_points(point_xy, polyline_3pts)
                if ind0 == 0: # left
                    point_projected = point_prj_0
                    ind_insert = ind_nearest
                elif ind0 == 2: #right
                    point_projected = point_prj_1
                    ind_insert = ind_nearest+1
                else:                    
                    ind_insert = ind_nearest
                    point_projected = polyline_3pts[ind0]                
        point_projected = point_projected.reshape(1, 2)
    else:
        raise IOError('polyline must contains more than one points')
    

    return point_projected, ind_insert

def point2segment(points, segment):
    """ project point(s) to a line segment and identify its position  
    points: nx2 array
    segment: 2X2 array
    -------
    point_prj : nx2 array
        coordinates of the projected point.
    position: True|False
        indicate whether the projected point is in/outside the line segment.
        True: inside, False: outside 
    """
    p0 = segment[0]
    p1 = segment[1]
    line_vect = np.array(p0)-np.array(p1)
    if norm(line_vect)==0:
        unit_vect = line_vect*0
    else:
        unit_vect = line_vect/norm(line_vect)
    dot_values = np.dot(points-p0, unit_vect)
    if dot_values.size > 1:
        point_prj = p0+unit_vect*dot_values[:, None]
    else:
        point_prj = p0+unit_vect*dot_values
        point_prj = point_prj.reshape(1,2)
    dis2end = np.c_[norm(point_prj-p0, axis=1), norm(point_prj-p1, axis=1)]
    
    position = np.max(dis2end, axis=1) <= norm(p1-p0)
    if points.size == 2:
        position = position[0]
    return point_prj, position

#%% supporting functions
def find_inter_point_on_polyline(polyline, straight_line, cmpr_dist=False, 
                                 avoid_none=False):
    """find the intersection point between a straigt line and a poly line
    If there is no intersection, project the near end point of polyline to 
    line section
    
    Parameters
    ----------
    polyline : 2-col numpy array
        points of a polyline.
    straight_line : 2x2 numpy array
        points of a line section(only two points).
    cmpr_dist: whether compare the distances between straight_line and the 
        nearest crossed-segment and the length of straight_line

    Returns
    -------
    inter_point.
    inter_ind: index on the first point of the crossed segment

    """
    line_p1 = straight_line[0]
    line_p2 = straight_line[-1]
    inter_ind = crossed_segments_on_polyline(polyline, line_p1, line_p2, 
                                             cmpr_dist)
    if inter_ind is None:
        inter_point = None
    else:
        segment_p0 = polyline[inter_ind]
        segment_p1 = polyline[inter_ind+1]
        segment = polyline[inter_ind:inter_ind+2]
        if norm(segment_p0-segment_p1) == 0:
            inter_point = segment_p0
        else:
            # lines1 = np.r_[line_p1.reshape((1,2)), line_p2.reshape((1,2))]
            inter_point = get_intersection(straight_line, segment)
    if avoid_none & (inter_ind is None): # choose the nearest point
        point_mid = np.mean(straight_line, axis=0)
        inter_ind = nearest_points(point_mid, polyline, 1)
        inter_point = polyline[inter_ind]
    return inter_point, inter_ind

def crossed_segments_on_polyline(polyline_xy, line_p1, line_p2, 
                                 cmpr_dist=False):
    # line_p1, line_p2: two points of a straight line
    # polyline_xy: xy coordinates of polyline
    # line_p1 = np.array(line_p1).reshape((1,2))
    # line_p2 = np.array(line_p2).reshape((1,2))
    segments_p0 = polyline_xy[:-1, :]
    segments_p1 = polyline_xy[1:, :]
    side_values0 = which_side_to_line(line_p1, line_p2, segments_p0)
    side_values1 = which_side_to_line(line_p1, line_p2, segments_p1)
    ind = np.where(side_values0*side_values1 <= 0)
    ind = ind[0]
    if ind.size>0:
        # fine the nearest segment
        line_mid = (line_p1+line_p2)/2
        segments_p0 = segments_p0[ind, :]
        segments_p1 = segments_p1[ind, :]
        segments_mid = (segments_p0+segments_p1)/2
        distance = norm(segments_mid-line_mid, axis=1)
        ind_sort = distance.argsort()
        ind_crossed = ind[ind_sort[0]]
        if cmpr_dist:
            # distance between straight line midpoint and polyline_xy
            cl = np.r_[line_p1.reshape(1,2), line_p2.reshape(1,2)]
            s_crssd = polyline_xy[ind_crossed:ind_crossed+2]
            inter_point = get_intersection(cl, s_crssd)
            d2bline = norm(inter_point-line_mid)
            # length of the straight line
            d_cline = norm(line_p1-line_p2)
            if d2bline > d_cline: 
                #it means the crossed segment is far away, the crossline does not
                # cross banline in the nearest position
                ind_crossed = None
    else:
        ind_crossed = None
    return ind_crossed

def which_side_to_line(line_p1, line_p2, points_xy):
    A, B, C = get_line_coefs(line_p1, line_p2)
    # A * x + B * y = C
    x = points_xy[:, 0]
    y = points_xy[:, 1]
    values = A*x + B*y - C
    values[values < 0] = -1
    values[values > 0] = 1
    return values
    
def get_line_coefs(p1, p2):
    """ Get line coefs of two points
    p1/p2: [2*N array] x and y coordinates 
    A * x + B * y = C
    """
    A = (p1[1] - p2[1])
    B = (p2[0] - p1[0])
    C = (p2[0]*p1[1] - p1[0]*p2[1])
    return A, B, C

def get_intersection(lines1, lines2):
    """ Get the intersection of two straight line or lines 
    lines1/lines2: 2*2*N array, 
        axis 0: point 1~2
        axis 1: x, y
        axis 2: line 1~N
    if lines2 contains only one line while lines1 contains more, then lines2 
        will be duplicated to match the size of lines1
    # A1 * x + B1 * y = C1
    # A2 * x + B2 * y = C2
    """
    # for line 1, 2xN array
    p1 = lines1[0, 0:2] # first point of line
    p2 = lines1[1, 0:2] # second point of line
    L1 = get_line_coefs(p1, p2)
    # for line 2
    p1 = lines2[0, 0:2]
    p2 = lines2[1, 0:2] 
    L2 = get_line_coefs(p1, p2)
    D  = L1[0] * L2[1] - L1[1] * L2[0]
    Dx = L1[2] * L2[1] - L1[1] * L2[2]
    Dy = L1[0] * L2[2] - L1[2] * L2[0]
    D = D.flatten()
    D[D==0] = np.nan
    x = Dx / D
    y = Dy / D
    if np.array([x, y]).size == 2:
        inter_point = np.array([x, y]).reshape(1, 2)
    else:
        inter_point = (x, y)
    return inter_point

def nearest_points(point_xy, poly_points, num=1):
    """
    find the nearest point on a line to a given point 
    Returns
    -------
    ind : index of the first nearest point.
    """
    distance = norm(poly_points-point_xy, axis=1)
    ind_sort = distance.argsort()
    if num==1:
        ind_nearest = ind_sort[0]
    else:
        ind_nearest = ind_sort[:num]
    return ind_nearest

def in_channel(channel_dict, xy_points, radius=-0.1):
    if 'polygon' in channel_dict.keys():
        poly_points = channel_dict['polygon']
    else:
        bank0 = channel_dict['bank0']
        bank1 = np.flipud(channel_dict['bank1'])
        cross0 = np.flipud(channel_dict['cross0'])
        cross1 = channel_dict['cross1']
        poly_points = np.r_[bank0, cross1[:, :2], bank1, cross0[:, :2]]
    section_poly = path.Path(poly_points)
    ind = section_poly.contains_points(xy_points, radius=radius)
    return ind

def distance_p2l(line_straight, points):
    """Calculate the distance from a point to a line defined by two points
    line_a: 2x2 array, 1st col is X and 2nd col is Y
    points: x and y coordinates of a point, can also be a 2-col array for 
        more than one points
    """
    p1 = line_straight[0, 0:2]
    p2 = line_straight[1, 0:2]
    if norm(p2-p1)==0:
        distance = norm(points-p1, axis=1)
    else:
        distance = np.cross(p2-p1, points-p1, axisb=1)/norm(p2-p1)
    distance = np.abs(distance)
    return distance

def remove_duplicate_rows(data_array):
    """ Remove duplicated points and preserve its order
    """
    u, ind = np.unique(data_array, axis=0, return_index=True)
    array_u = u[np.argsort(ind)]
    return array_u

def within_segment(point_xy, line_segment):
    dis2end = [norm(point_xy-line_segment[0]), 
               norm(point_xy-line_segment[1])]
    if np.max(dis2end)>norm(line_segment[0]-line_segment[1]):
        ind_logical = False
    else:
        ind_logical = True
    return ind_logical

def get_bounding_box(river_channel, margin=0.1):
    if hasattr(river_channel, 'get_extents'):
        xy = river_channel.get_extents().get_points()
    else:
        xy = np.array(river_channel)
    x_min = np.min(xy[:, 0])
    x_max = np.max(xy[:, 0])
    x_range = x_max-x_min
    y_min = np.min(xy[:, 1])
    y_max = np.max(xy[:, 1])
    y_range = y_max-y_min
    left = x_min-margin*x_range
    right = x_max+margin*x_range
    bottom = y_min-margin*y_range
    top = y_max+margin*y_range
    bounding_box = np.array([[left, bottom], [right, top]] )
    return bounding_box
    
    
    
    